import asyncio
import logging
import os
from pathlib import Path
from unittest.mock import ANY

import pytest

from ophyd_async.core import (
    Device,
    HDFDocumentComposer,
    SignalR,
    StaticFilenameProvider,
    StaticPathProvider,
    init_devices,
)
from ophyd_async.fastcs.core import fastcs_connector
from ophyd_async.fastcs.panda import (
    CommonPandaBlocks,
    DatasetTable,
    PandaHdf5DatasetType,
    PandaHDFWriter,
)
from ophyd_async.testing import callback_on_mock_put, set_mock_value

PANDA_DETECTOR_NAME = "mock_panda"

TABLES = [
    DatasetTable(
        name=[],
        dtype=[],
    ),
    DatasetTable(
        name=["x"],
        dtype=[PandaHdf5DatasetType.UINT_32],
    ),
    DatasetTable(
        name=[
            "x",
            "y",
            "y_min",
            "y_max",
        ],
        dtype=[
            PandaHdf5DatasetType.UINT_32,
            PandaHdf5DatasetType.FLOAT_64,
            PandaHdf5DatasetType.FLOAT_64,
            PandaHdf5DatasetType.FLOAT_64,
        ],
    ),
]


@pytest.fixture
async def panda_t():
    class CaptureBlock(Device):
        test_capture: SignalR[float]

    class Panda(CommonPandaBlocks):
        block_a: CaptureBlock
        block_b: CaptureBlock

        def __init__(self, uri: str, name: str = ""):
            super().__init__(name=name, connector=fastcs_connector(self, uri))

    yield Panda


@pytest.fixture
async def mock_panda(panda_t):
    async with init_devices(mock=True):
        mock_panda = panda_t("mock_PANDA", name="mock_panda")

    # Mimic directory exists check that happens normally in the PandA IOC
    def check_dir_exits(value, **kwargs):
        if os.path.exists(os.path.abspath(os.path.dirname(value))):
            set_mock_value(mock_panda.data.directory_exists, 1)

    # Assume directory exists
    callback_on_mock_put(mock_panda.data.hdf_directory, check_dir_exits)

    set_mock_value(
        mock_panda.data.datasets,
        DatasetTable(
            name=[],
            dtype=[],
        ),
    )

    return mock_panda


@pytest.fixture
async def mock_writer(tmp_path, mock_panda) -> PandaHDFWriter:
    fp = StaticFilenameProvider("data")
    dp = StaticPathProvider(fp, tmp_path / mock_panda.name, create_dir_depth=-1)
    async with init_devices(mock=True):
        writer = PandaHDFWriter(
            path_provider=dp,
            panda_data_block=mock_panda.data,
        )

    return writer


@pytest.mark.parametrize("table", TABLES)
async def test_open_returns_correct_descriptors(
    mock_writer: PandaHDFWriter, table: DatasetTable, caplog
):
    assert hasattr(mock_writer, "panda_data_block")
    set_mock_value(
        mock_writer.panda_data_block.datasets,
        table,
    )

    with caplog.at_level(logging.WARNING):
        description = await mock_writer.open(
            PANDA_DETECTOR_NAME
        )  # to make capturing status not time out

        # Check if empty datasets table leads to warning log message
        if len(table.name) == 0:
            assert "DATASETS table is empty!" in caplog.text

    for key, entry, expected_key in zip(
        description.keys(), description.values(), table.name, strict=False
    ):
        assert key == expected_key
        assert entry == {
            "source": mock_writer.panda_data_block.hdf_directory.source,
            "shape": [],
            "dtype": "number",
            "dtype_numpy": "<f8",
            "external": "STREAM:",
        }


async def test_open_close_sets_capture(mock_writer: PandaHDFWriter):
    assert isinstance(await mock_writer.open(PANDA_DETECTOR_NAME), dict)
    assert await mock_writer.panda_data_block.capture.get_value()
    await mock_writer.close()
    assert not await mock_writer.panda_data_block.capture.get_value()


async def test_open_sets_file_path_and_name(mock_writer: PandaHDFWriter, tmp_path):
    await mock_writer.open(PANDA_DETECTOR_NAME)
    path = await mock_writer.panda_data_block.hdf_directory.get_value()
    assert path == str(tmp_path / PANDA_DETECTOR_NAME)
    name = await mock_writer.panda_data_block.hdf_file_name.get_value()
    assert name == "data.h5"


async def test_get_indices_written(mock_writer: PandaHDFWriter):
    await mock_writer.open(PANDA_DETECTOR_NAME)
    set_mock_value(mock_writer.panda_data_block.num_captured, 4)
    written = await mock_writer.get_indices_written()
    assert written == 4


@pytest.mark.parametrize("exposures_per_event", [1, 2, 11])
async def test_wait_for_index(mock_writer: PandaHDFWriter, exposures_per_event: int):
    await mock_writer.open(PANDA_DETECTOR_NAME, exposures_per_event=exposures_per_event)
    set_mock_value(mock_writer.panda_data_block.num_captured, 3 * exposures_per_event)
    await mock_writer.wait_for_index(3, timeout=1)
    set_mock_value(mock_writer.panda_data_block.num_captured, 2 * exposures_per_event)
    with pytest.raises(asyncio.TimeoutError):
        await mock_writer.wait_for_index(3, timeout=0.1)


@pytest.mark.parametrize("table", TABLES)
async def test_collect_stream_docs(
    mock_writer: PandaHDFWriter,
    tmp_path: Path,
    table: DatasetTable,
):
    # Give the mock writer datasets
    set_mock_value(mock_writer.panda_data_block.datasets, table)

    await mock_writer.open(PANDA_DETECTOR_NAME)

    def assert_resource_document(name, resource_doc):
        assert resource_doc == {
            "uid": ANY,
            "data_key": name,
            "mimetype": "application/x-hdf5",
            "uri": "file://localhost/"
            + (tmp_path / "mock_panda" / "data.h5").as_posix().lstrip("/"),
            "parameters": {
                "dataset": f"/{name}",
                "chunk_shape": (1024,),
            },
        }

        # URI will always use '/'
        assert "mock_panda/data.h5" in resource_doc["uri"]

    [item async for item in mock_writer.collect_stream_docs(PANDA_DETECTOR_NAME, 1)]
    assert type(mock_writer._composer) is HDFDocumentComposer
    assert mock_writer._composer._last_emitted == 1

    for i in range(len(table.name)):
        resource_doc = mock_writer._composer._bundles[i].stream_resource_doc
        name = table.name[i]

        assert_resource_document(name=name, resource_doc=resource_doc)

        assert resource_doc["data_key"] == name


async def test_oserror_when_hdf_dir_does_not_exist(tmp_path, mock_panda):
    fp = StaticFilenameProvider("data")
    dp = StaticPathProvider(
        fp, tmp_path / mock_panda.name / "extra" / "dirs", create_dir_depth=-1
    )
    async with init_devices(mock=True):
        writer = PandaHDFWriter(
            path_provider=dp,
            panda_data_block=mock_panda.data,
        )

    with pytest.raises(OSError):
        await writer.open(PANDA_DETECTOR_NAME)



from collections.abc import Sequence

from ophyd_async.core import PathProvider, SignalR
from ophyd_async.epics.adcore import (
    ADHDFWriter,
    ADWriter,
    AreaDetector,
    NDPluginBaseIO,
)

from ._kinetix_controller import KinetixController
from ._kinetix_io import KinetixDriverIO


class KinetixDetector(AreaDetector[KinetixController]):
    """Ophyd-async implementation of an ADKinetix Detector.

    https://github.com/NSLS-II/ADKinetix.
    """

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider,
        drv_suffix: str = "cam1:",
        writer_cls: type[ADWriter] = ADHDFWriter,
        fileio_suffix: str | None = None,
        name: str = "",
        plugins: dict[str, NDPluginBaseIO] | None = None,
        config_sigs: Sequence[SignalR] = (),
    ):
        driver = KinetixDriverIO(prefix + drv_suffix)
        controller = KinetixController(driver)

        writer = writer_cls.with_io(
            prefix,
            path_provider,
            dataset_source=driver,
            fileio_suffix=fileio_suffix,
            plugins=plugins,
        )

        super().__init__(
            controller=controller,
            writer=writer,
            plugins=plugins,
            name=name,
            config_sigs=config_sigs,
        )



from abc import ABC, abstractmethod
from typing import Any, Generic

from bluesky.protocols import Flyable, Preparable, Stageable
from pydantic import Field

from ._device import Device
from ._status import AsyncStatus
from ._utils import CALCULATE_TIMEOUT, CalculatableTimeout, ConfinedModel, T


class FlyerController(ABC, Generic[T]):
    """Base class for controlling 'flyable' devices.

    [`bluesky.protocols.Flyable`](#bluesky.protocols.Flyable).
    """

    @abstractmethod
    async def prepare(self, value: T) -> Any:
        """Move to the start of the fly scan."""

    @abstractmethod
    async def kickoff(self):
        """Start the fly scan."""

    @abstractmethod
    async def complete(self):
        """Block until the fly scan is done."""

    @abstractmethod
    async def stop(self):
        """Stop flying and wait everything to be stopped."""


class FlyMotorInfo(ConfinedModel):
    """Minimal set of information required to fly a motor."""

    start_position: float = Field(frozen=True)
    """Absolute position of the motor once it finishes accelerating to desired
    velocity, in motor EGUs"""

    end_position: float = Field(frozen=True)
    """Absolute position of the motor once it begins decelerating from desired
    velocity, in EGUs"""

    time_for_move: float = Field(frozen=True, gt=0)
    """Time taken for the motor to get from start_position to end_position, excluding
    run-up and run-down, in seconds."""

    timeout: CalculatableTimeout = Field(frozen=True, default=CALCULATE_TIMEOUT)
    """Maximum time for the complete motor move, including run up and run down.
    Defaults to `time_for_move` + run up and run down times + 10s."""

    @property
    def velocity(self) -> float:
        """Calculate the velocity of the constant velocity phase."""
        return (self.end_position - self.start_position) / self.time_for_move

    def ramp_up_start_pos(self, acceleration_time: float) -> float:
        """Calculate the start position with run-up distance added on."""
        return self.start_position - acceleration_time * self.velocity / 2

    def ramp_down_end_pos(self, acceleration_time: float) -> float:
        """Calculate the end position with run-down distance added on."""
        return self.end_position + acceleration_time * self.velocity / 2


class StandardFlyer(
    Device,
    Stageable,
    Preparable,
    Flyable,
    Generic[T],
):
    """Base class for 'flyable' devices.

    [`bluesky.protocols.Flyable`](#bluesky.protocols.Flyable).
    """

    def __init__(
        self,
        trigger_logic: FlyerController[T],
        name: str = "",
    ):
        self._trigger_logic = trigger_logic
        super().__init__(name=name)

    @property
    def trigger_logic(self) -> FlyerController[T]:
        return self._trigger_logic

    @AsyncStatus.wrap
    async def stage(self) -> None:
        await self.unstage()

    @AsyncStatus.wrap
    async def unstage(self) -> None:
        await self._trigger_logic.stop()

    def prepare(self, value: T) -> AsyncStatus:
        return AsyncStatus(self._prepare(value))

    async def _prepare(self, value: T) -> None:
        # Move to start and setup the fly scan
        await self._trigger_logic.prepare(value)

    @AsyncStatus.wrap
    async def kickoff(self) -> None:
        await self._trigger_logic.kickoff()

    @AsyncStatus.wrap
    async def complete(self) -> None:
        await self._trigger_logic.complete()



import os
import sys
from pathlib import Path
from typing import Any
from unittest.mock import ANY, patch

import bluesky.plan_stubs as bps
import bluesky.plans as bp
import pytest
from bluesky.run_engine import RunEngine
from bluesky.utils import MsgGenerator
from event_model.documents import (
    DocumentType,
    Event,
    EventDescriptor,
    RunStart,
    RunStop,
    StreamDatum,
    StreamResource,
)

from ophyd_async.core import (
    StaticPathProvider,
    TriggerInfo,
    YamlSettingsProvider,
    init_devices,
)
from ophyd_async.epics.adsimdetector import SimDetector
from ophyd_async.plan_stubs import (
    apply_settings,
    apply_settings_if_different,
    get_current_settings,
    retrieve_settings,
)

TIMEOUT = 10.0


@pytest.fixture(scope="module", autouse=True)
def with_env():
    with patch.dict(
        os.environ,
        {
            "EPICS_CA_NAME_SERVERS": "127.0.0.1:9064",
            "EPICS_PVA_NAME_SERVERS": "127.0.0.1:9075",
        },
        clear=False,
    ):
        yield


@pytest.fixture
def adsim(RE: RunEngine) -> SimDetector:
    prefix = "BL01T"
    provider = StaticPathProvider(lambda _: "adsim", Path("/tmp"))
    with init_devices():
        adsim = SimDetector(
            f"{prefix}-DI-CAM-01:",
            path_provider=provider,
            drv_suffix="DET:",
            fileio_suffix="HDF5:",
        )

    RE(apply_baseline_settings(adsim))

    return adsim


def apply_baseline_settings(adsim: SimDetector) -> MsgGenerator[None]:
    current_settings = yield from get_current_settings(adsim)
    provider = YamlSettingsProvider(Path(__file__).parent)
    baseline_settings = yield from retrieve_settings(
        provider,
        "baseline",
        adsim,
    )
    yield from apply_settings_if_different(
        baseline_settings,
        apply_plan=apply_settings,
        current_settings=current_settings,
    )


@pytest.mark.timeout(TIMEOUT + 3.0)
@pytest.mark.xfail(reason="https://github.com/bluesky/ophyd-async/issues/998")
def test_prepare_is_idempotent_and_sets_exposure_time(
    RE: RunEngine, adsim: SimDetector
) -> None:
    def prepare_then_count() -> MsgGenerator[None]:
        yield from bps.prepare(
            adsim,
            TriggerInfo(livetime=0.2),
            wait=True,
        )
        yield from bp.count([adsim])

    RE(prepare_then_count())

    actual_exposure_time: float = RE(bps.rd(adsim.driver.acquire_time)).plan_result
    assert actual_exposure_time == 0.2


@pytest.mark.insubprocess
@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Services not set up on Windows"
)
@pytest.mark.timeout(TIMEOUT + 15.0)
def test_software_triggering(RE: RunEngine, adsim: SimDetector) -> None:
    docs = run_plan_and_get_documents(RE, bp.count([adsim], num=2))
    assert docs == [
        RunStart(
            uid=ANY,
            time=ANY,
            versions=ANY,
            scan_id=ANY,
            plan_type="generator",
            plan_name="count",
            detectors=["adsim"],
            num_points=2,
            num_intervals=1,
            plan_args={
                "detectors": [ANY],
                "num": 2,
                "delay": 0.0,
            },
            hints={
                "dimensions": [
                    (
                        ("time",),
                        "primary",
                    ),
                ],
            },
        ),
        EventDescriptor(
            uid=ANY,
            time=ANY,
            name="primary",
            configuration={
                "adsim": {
                    "data": {
                        "adsim-driver-acquire_period": 0.005,
                        "adsim-driver-acquire_time": 0.1,
                    },
                    "timestamps": {
                        "adsim-driver-acquire_period": ANY,
                        "adsim-driver-acquire_time": ANY,
                    },
                    "data_keys": {
                        "adsim-driver-acquire_period": {
                            "dtype": "number",
                            "shape": [],
                            "dtype_numpy": "<f8",
                            "source": "ca://BL01T-DI-CAM-01:DET:AcquirePeriod_RBV",
                            "units": "",
                            "precision": 3,
                        },
                        "adsim-driver-acquire_time": {
                            "dtype": "number",
                            "shape": [],
                            "dtype_numpy": "<f8",
                            "source": "ca://BL01T-DI-CAM-01:DET:AcquireTime_RBV",
                            "units": "",
                            "precision": 3,
                        },
                    },
                }
            },
            data_keys={
                "adsim": {
                    "source": "ca://BL01T-DI-CAM-01:HDF5:FullFileName_RBV",
                    "shape": [1, 1024, 1024],
                    "dtype": "array",
                    "dtype_numpy": "|i1",
                    "external": "STREAM:",
                    "object_name": "adsim",
                }
            },
            object_keys={"adsim": ["adsim"]},
            run_start=ANY,
            hints={"adsim": {"fields": ["adsim"]}},
        ),
        StreamResource(
            uid=ANY,
            run_start=ANY,
            data_key="adsim",
            mimetype="application/x-hdf5",
            uri="file://localhost/tmp/adsim.h5",
            parameters={
                "dataset": "/entry/data/data",
                "chunk_shape": (1, 1024, 1024),
            },
        ),
        StreamDatum(
            stream_resource=ANY,
            descriptor=ANY,
            uid=ANY,
            seq_nums={"start": 1, "stop": 2},
            indices={"start": 0, "stop": 1},
        ),
        Event(
            uid=ANY,
            time=ANY,
            descriptor=ANY,
            data={},
            timestamps={},
            seq_num=1,
            filled={},
        ),
        StreamDatum(
            stream_resource=ANY,
            descriptor=ANY,
            uid=ANY,
            seq_nums={"start": 2, "stop": 3},
            indices={"start": 1, "stop": 2},
        ),
        Event(
            uid=ANY,
            time=ANY,
            descriptor=ANY,
            data={},
            timestamps={},
            seq_num=2,
            filled={},
        ),
        RunStop(
            run_start=ANY,
            uid=ANY,
            time=ANY,
            exit_status="success",
            reason="",
            num_events={"primary": 2},
        ),
    ]


def run_plan_and_get_documents(
    RE: RunEngine, plan: MsgGenerator[Any]
) -> list[DocumentType]:
    docs = []
    RE(plan, lambda name, doc: docs.append(doc))
    return docs



import os

import pytest

from ophyd_async.core import (
    StaticFilenameProvider,
    StaticPathProvider,
)
from ophyd_async.fastcs.panda import (
    DatasetTable,
    HDFPanda,
    PandaHdf5DatasetType,
)
from ophyd_async.testing import callback_on_mock_put, set_mock_value


@pytest.fixture
async def mock_hdf_panda(tmp_path):
    fp = StaticFilenameProvider("test-panda")
    dp = StaticPathProvider(fp, tmp_path)

    mock_hdf_panda = HDFPanda("HDFPANDA:", path_provider=dp, name="panda")
    await mock_hdf_panda.connect(mock=True)

    def link_function(value: bool, wait: bool = True):
        set_mock_value(mock_hdf_panda.pcap.active, value)

    # Mimic directory exists check that happens normally in the PandA IOC
    def check_dir_exits(value: str, wait: bool = True):
        if os.path.exists(value):
            set_mock_value(mock_hdf_panda.data.directory_exists, True)

    callback_on_mock_put(mock_hdf_panda.pcap.arm, link_function)
    callback_on_mock_put(mock_hdf_panda.data.hdf_directory, check_dir_exits)

    set_mock_value(
        mock_hdf_panda.data.datasets,
        DatasetTable(
            name=["x", "y"],
            dtype=[PandaHdf5DatasetType.UINT_32, PandaHdf5DatasetType.FLOAT_64],
        ),
    )

    yield mock_hdf_panda


async def test_hdf_panda_passes_blocks_to_controller(mock_hdf_panda: HDFPanda):
    assert hasattr(mock_hdf_panda._controller, "pcap")
    assert mock_hdf_panda._controller.pcap is mock_hdf_panda.pcap



from unittest.mock import ANY, call

from pytest import fixture

from ophyd_async.core import (
    TriggerInfo,
    init_devices,
)
from ophyd_async.fastcs.eiger import EigerController, EigerDriverIO
from ophyd_async.testing import (
    callback_on_mock_put,
    get_mock,
    get_mock_put,
    set_mock_value,
)

DriverAndController = tuple[EigerDriverIO, EigerController]


@fixture
def eiger_driver_and_controller_no_arm(RE) -> DriverAndController:
    with init_devices(mock=True):
        driver = EigerDriverIO("")
        controller = EigerController(driver)

    def become_idle_after_arm(*args, **kwargs):
        # Mocking that eiger has armed and finished taking frames.
        set_mock_value(driver.detector.state, "idle")

    callback_on_mock_put(driver.detector.arm, become_idle_after_arm)

    return driver, controller


@fixture
def eiger_driver_and_controller(
    eiger_driver_and_controller_no_arm: DriverAndController,
) -> DriverAndController:
    driver, controller = eiger_driver_and_controller_no_arm

    return driver, controller


async def test_when_arm_with_exposure_then_time_and_period_set(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller
    test_exposure = 0.002
    await controller.prepare(TriggerInfo(number_of_events=10, livetime=test_exposure))
    await controller.arm()
    await controller.wait_for_idle()
    assert (await driver.detector.frame_time.get_value()) == test_exposure
    assert (await driver.detector.count_time.get_value()) == test_exposure


async def test_when_arm_with_no_exposure_then_arm_set_correctly(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller
    await controller.prepare(TriggerInfo(number_of_events=10))
    await controller.arm()
    await controller.wait_for_idle()
    get_mock_put(driver.detector.arm).assert_called_once()


async def test_when_arm_with_number_of_images_then_number_of_images_set_correctly(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller
    test_number_of_images = 40
    await controller.prepare(TriggerInfo(number_of_events=test_number_of_images))
    await controller.arm()
    await controller.wait_for_idle()
    get_mock_put(driver.detector.nimages).assert_called_once_with(
        test_number_of_images, wait=ANY
    )


async def test_when_disarm_called_on_controller_then_disarm_called_on_driver(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller
    await controller.disarm()
    get_mock_put(driver.detector.disarm).assert_called_once()


async def test_when_get_deadtime_called_then_returns_expected_deadtime(
    eiger_driver_and_controller: DriverAndController,
):
    _, controller = eiger_driver_and_controller
    assert controller.get_deadtime(0) == 0.0001


async def test_given_energy_within_tolerance_when_photon_energy_set_then_pv_unchanged(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller
    initial_energy = 10
    set_mock_value(driver.detector.photon_energy, initial_energy)
    await controller.set_energy(10.002)
    get_mock_put(driver.detector.photon_energy).assert_not_called()
    assert (await driver.detector.photon_energy.get_value()) == initial_energy


async def test_given_energy_outside_tolerance_when_photon_energy_set_then_pv_changed(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller
    initial_energy = 10
    new_energy = 15
    set_mock_value(driver.detector.photon_energy, initial_energy)
    await controller.set_energy(new_energy)
    get_mock_put(driver.detector.photon_energy).assert_called_once()
    assert (await driver.detector.photon_energy.get_value()) == new_energy


async def test_when_prepare_called__correct_parameters_set(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller

    await controller.prepare(TriggerInfo(livetime=1))

    detector_mock = get_mock(driver.detector)
    mock_calls = detector_mock.mock_calls
    assert [
        call.trigger_mode.put("ints", wait=True),
        call.nimages.put(1, wait=True),
        call.count_time.put(1.0, wait=True),
        call.frame_time.put(1.0, wait=True),
    ] in mock_calls



from typing import Annotated as A

from ophyd_async.core import DEFAULT_TIMEOUT, AsyncStatus, SignalR, SignalRW, SignalX
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.tango.core import TangoPolling, TangoReadable


class TangoCounter(TangoReadable):
    """Tango counting device."""

    # Enter the name and type of the signals you want to use
    # If the server doesn't support events, the TangoPolling annotation gives
    # the parameters for ophyd to poll instead
    counts: A[SignalR[int], Format.HINTED_SIGNAL, TangoPolling(1.0, 0.1, 0.1)]
    sample_time: A[SignalRW[float], Format.CONFIG_SIGNAL, TangoPolling(0.1, 0.1, 0.1)]
    start: SignalX
    # If a tango name clashes with a bluesky verb, add a trailing underscore
    reset_: SignalX

    @AsyncStatus.wrap
    async def trigger(self) -> None:
        sample_time = await self.sample_time.get_value()
        timeout = sample_time + DEFAULT_TIMEOUT
        await self.start.trigger(wait=True, timeout=timeout)

    @AsyncStatus.wrap
    async def reset(self) -> None:
        await self.reset_.trigger(wait=True, timeout=DEFAULT_TIMEOUT)



from __future__ import annotations

import asyncio
import sys
from collections.abc import Awaitable, Callable, Iterator, Mapping, MutableMapping
from functools import cached_property
from logging import LoggerAdapter, getLogger
from typing import Any, TypeVar

from bluesky.protocols import HasName
from bluesky.run_engine import call_in_bluesky_event_loop, in_bluesky_event_loop

from ._utils import (
    DEFAULT_TIMEOUT,
    LazyMock,
    NotConnectedError,
    error_if_none,
    wait_for_connection,
)


class DeviceConnector:
    """Defines how a `Device` should be connected and type hints processed."""

    def create_children_from_annotations(self, device: Device):
        """Use when children can be created from introspecting the hardware.

        Some control systems allow introspection of a device to determine what
        children it has. To allow this to work nicely with typing we add these
        hints to the Device like so::

            my_signal: SignalRW[int]
            my_device: MyDevice

        This method will be run during `Device.__init__`, and is responsible
        for turning all of those type hints into real Signal and Device instances.

        Subsequent runs of this function should do nothing, to allow it to be
        called early in Devices that need to pass references to their children
        during `__init__`.
        """

    async def connect_mock(self, device: Device, mock: LazyMock):
        """Use during [](#Device.connect) with `mock=True`.

        This is called when there is no cached connect done in `mock=True`
        mode. It connects the Device and all its children in mock mode.
        """
        # Connect serially, no errors to gather up as in mock mode
        exceptions: dict[str, Exception] = {}
        for name, child_device in device.children():
            try:
                await child_device.connect(mock=mock.child(name))
            except Exception as exc:
                exceptions[name] = exc
        if exceptions:
            raise NotConnectedError.with_other_exceptions_logged(exceptions)

    async def connect_real(self, device: Device, timeout: float, force_reconnect: bool):
        """Use during [](#Device.connect) with `mock=False`.

        This is called when there is no cached connect done in `mock=False`
        mode. It connects the Device and all its children in real mode in parallel.
        """
        # Connect in parallel, gathering up NotConnectedErrors
        coros = {
            name: child_device.connect(timeout=timeout, force_reconnect=force_reconnect)
            for name, child_device in device.children()
        }
        await wait_for_connection(**coros)


class Device(HasName):
    """Common base class for all Ophyd Async Devices.

    :param name: Optional name of the Device
    :param connector: Optional DeviceConnector instance to use at connect()
    """

    parent: Device | None = None
    """The parent Device if it exists"""
    _name: str = ""
    # None if connect hasn't started, a Task if it has
    _connect_task: asyncio.Task | None = None
    # The mock if we have connected in mock mode
    _mock: LazyMock | None = None
    # The separator to use when making child names
    _child_name_separator: str = "-"

    def __init__(
        self, name: str = "", connector: DeviceConnector | None = None
    ) -> None:
        self._connector = connector or DeviceConnector()
        self._connector.create_children_from_annotations(self)
        if name:
            self.set_name(name)

    @property
    def name(self) -> str:
        """Return the name of the Device."""
        return self._name

    @cached_property
    def _child_devices(self) -> dict[str, Device]:
        return {}

    def children(self) -> Iterator[tuple[str, Device]]:
        """For each attribute that is a Device, yield the name and Device.

        :yields: `(attr_name, attr)` for each child attribute that is a Device.
        """
        yield from self._child_devices.items()

    @cached_property
    def log(self) -> LoggerAdapter:
        """Return a logger configured with the device name."""
        return LoggerAdapter(
            getLogger("ophyd_async.devices"), {"ophyd_async_device_name": self.name}
        )

    def set_name(self, name: str, *, child_name_separator: str | None = None) -> None:
        """Set `self.name=name` and each `self.child.name=name+"-child"`.

        :param name: New name to set.
        :param child_name_separator:
            Use this as a separator instead of "-". Use "_" instead to make the
            same names as the equivalent ophyd sync device.
        """
        self._name = name
        if child_name_separator:
            self._child_name_separator = child_name_separator
        # Ensure logger is recreated after a name change
        if "log" in self.__dict__:
            del self.log
        for attr_name, child in self.children():
            child_name = (
                f"{self.name}{self._child_name_separator}{attr_name}"
                if self.name
                else ""
            )
            child.set_name(child_name, child_name_separator=self._child_name_separator)

    def __setattr__(self, name: str, value: Any) -> None:
        # Bear in mind that this function is called *a lot*, so
        # we need to make sure nothing expensive happens in it...
        if name == "parent":
            if self.parent not in (value, None):
                raise TypeError(
                    f"Cannot set the parent of {self} to be {value}: "
                    f"it is already a child of {self.parent}"
                )
        # ...hence not doing an isinstance check for attributes we
        # know not to be Devices
        elif name not in _not_device_attrs and isinstance(value, Device):
            value.parent = self
            self._child_devices[name] = value
            # And if the name is set, then set the name of all children,
            # including the child
            if self._name:
                self.set_name(self._name)
        # ...and avoiding the super call as we know it resolves to `object`
        return object.__setattr__(self, name, value)

    async def connect(
        self,
        mock: bool | LazyMock = False,
        timeout: float = DEFAULT_TIMEOUT,
        force_reconnect: bool = False,
    ) -> None:
        """Connect the device and all child devices.

        Successful connects will be cached so subsequent calls will return
        immediately. Contains a timeout that gets propagated to child.connect
        methods.

        :param mock:
            If True then use [](#MockSignalBackend) for all Signals. If passed a
            [](#LazyMock) then pass this down for use within the Signals,
            otherwise create one.
        :param timeout: Time to wait before failing with a TimeoutError.
        :param force_reconnect:
            If True, force a reconnect even if the last connect succeeded.
        """
        connector = error_if_none(
            getattr(self, "_connector", None),
            f"{self}: doesn't have attribute `_connector`,"
            f" did you call `super().__init__` in your `__init__` method?",
        )
        if mock:
            # Always connect in mock mode serially
            if isinstance(mock, LazyMock):
                # Use the provided mock
                self._mock = mock
            elif not self._mock:
                # Make one
                self._mock = LazyMock()
            await connector.connect_mock(self, self._mock)
        else:
            # Try to cache the connect in real mode
            can_use_previous_connect = (
                self._mock is None
                and self._connect_task
                and not (self._connect_task.done() and self._connect_task.exception())
            )
            if force_reconnect or not can_use_previous_connect:
                self._mock = None
                coro = connector.connect_real(self, timeout, force_reconnect)
                self._connect_task = asyncio.create_task(coro)
            connect_task = error_if_none(
                self._connect_task, "Connect task not created, this shouldn't happen"
            )
            # Wait for it to complete
            await connect_task


_not_device_attrs = {
    "_name",
    "_children",
    "_connector",
    "_timeout",
    "_mock",
    "_connect_task",
}


DeviceT = TypeVar("DeviceT", bound=Device)


class DeviceVector(MutableMapping[int, DeviceT], Device):
    """Defines a dictionary of Device children with arbitrary integer keys.

    :see-also: [](#implementing-devices) for examples of how to use this class.
    """

    def __init__(
        self,
        children: Mapping[int, DeviceT],
        name: str = "",
    ) -> None:
        self._children: dict[int, DeviceT] = {}
        self.update(children)
        super().__init__(name=name)

    def __setattr__(self, name: str, child: Any) -> None:
        if name != "parent" and isinstance(child, Device):
            raise AttributeError(
                "DeviceVector can only have integer named children, "
                "set via device_vector[i] = child"
            )
        super().__setattr__(name, child)

    def __getitem__(self, key: int) -> DeviceT:
        return self._children[key]

    def __setitem__(self, key: int, value: DeviceT) -> None:
        # Check the types on entry to dict to make sure we can't accidentally
        # make a non-integer named child
        if not isinstance(key, int):
            msg = f"Expected int, got {key}"
            raise TypeError(msg)
        if not isinstance(value, Device):
            msg = f"Expected Device, got {value}"
            raise TypeError(msg)
        self._children[key] = value
        value.parent = self

    def __delitem__(self, key: int) -> None:
        del self._children[key]

    def __iter__(self) -> Iterator[int]:
        yield from self._children

    def __len__(self) -> int:
        return len(self._children)

    def children(self) -> Iterator[tuple[str, Device]]:
        for key, child in self._children.items():
            yield str(key), child

    def __hash__(self):  # to allow DeviceVector to be used as dict keys and in sets
        return hash(id(self))


class DeviceProcessor:
    """Sync/Async Context Manager that finds all the Devices declared within it.

    Used in `init_devices`
    """

    def __init__(self, process_devices: Callable[[dict[str, Device]], Awaitable[None]]):
        self._process_devices = process_devices
        self._locals_on_enter: dict[str, Any] = {}
        self._locals_on_exit: dict[str, Any] = {}

    def _caller_locals(self) -> dict[str, Any]:
        """Walk up until we find a stack frame that doesn't have us as self."""
        try:
            raise ValueError
        except ValueError:
            _, _, tb = sys.exc_info()
            tb = error_if_none(tb, "Can't get traceback, this shouldn't happen")

            caller_frame = tb.tb_frame
            while caller_frame.f_locals.get("self", None) is self:
                caller_frame = caller_frame.f_back
                if not caller_frame:
                    msg = (
                        "No previous frame to the one with self in it, "
                        "this shouldn't happen"
                    )
                    raise RuntimeError(  # noqa: B904
                        msg
                    )
            return caller_frame.f_locals.copy()

    def __enter__(self) -> DeviceProcessor:
        # Stash the names that were defined before we were called
        self._locals_on_enter = self._caller_locals()
        return self

    async def __aenter__(self) -> DeviceProcessor:
        return self.__enter__()

    async def __aexit__(self, type, value, traceback):
        self._locals_on_exit = self._caller_locals()
        await self._on_exit()

    def __exit__(self, type_, value, traceback):
        if in_bluesky_event_loop():
            raise RuntimeError(
                "Cannot use DeviceConnector inside a plan, instead use "
                "`yield from ophyd_async.plan_stubs.ensure_connected(device)`"
            )
        self._locals_on_exit = self._caller_locals()
        try:
            fut = call_in_bluesky_event_loop(self._on_exit())
        except RuntimeError as exc:
            raise NotConnectedError(
                "Could not connect devices. Is the bluesky event loop running? See "
                "https://blueskyproject.io/ophyd-async/main/"
                "user/explanations/event-loop-choice.html for more info."
            ) from exc
        return fut

    async def _on_exit(self) -> None:
        # Find all the devices
        devices = {
            name: obj
            for name, obj in self._locals_on_exit.items()
            if isinstance(obj, Device) and self._locals_on_enter.get(name) is not obj
        }
        # Call the provided process function on them
        await self._process_devices(devices)


def init_devices(
    set_name: bool = True,
    child_name_separator: str = "-",
    connect: bool = True,
    mock: bool = False,
    timeout: float = 10.0,
):
    """Auto initialize top level Device instances: to be used as a context manager.

    :param set_name:
        If True, call `device.set_name(variable_name)` on all Devices created
        within the context manager that have an empty `name`.
    :param child_name_separator: Separator for child names if `set_name` is True.
    :param connect:
        If True, call `device.connect(mock, timeout)` in parallel on all Devices
        created within the context manager.
    :param mock: If True, connect Signals in mock mode.
    :param timeout: How long to wait for connect before logging an exception.
    :raises RuntimeError: If used inside a plan, use [](#ensure_connected) instead.
    :raises NotConnectedError: If devices could not be connected.

    For example, to connect and name 2 motors in parallel:
    ```python
    [async] with init_devices():
        t1x = motor.Motor("BLxxI-MO-TABLE-01:X")
        t1y = motor.Motor("pva://BLxxI-MO-TABLE-01:Y")
        # Names and connects devices here
    assert t1x.name == "t1x"
    ```
    """

    async def process_devices(devices: dict[str, Device]):
        if set_name:
            for name, device in devices.items():
                if not device.name:
                    device.set_name(name, child_name_separator=child_name_separator)
        if connect:
            coros = {
                name: device.connect(mock, timeout) for name, device in devices.items()
            }
            await wait_for_connection(**coros)

    return DeviceProcessor(process_devices)



from __future__ import annotations

import asyncio

from ophyd_async.core import (
    Device,
    DeviceVector,
    StandardReadable,
    StrictEnum,
    derived_signal_r,
    derived_signal_rw,
    derived_signal_w,
    soft_signal_rw,
)


class BeamstopPosition(StrictEnum):
    IN_POSITION = "In position"
    OUT_OF_POSITION = "Out of position"


class ReadOnlyBeamstop(Device):
    """Reads from 2 motors to work out if the beamstop is in position.

    E.g. bps.rd(beamstop.position)
    """

    def __init__(self, name=""):
        # Raw signals
        self.x = soft_signal_rw(float)
        self.y = soft_signal_rw(float)
        # Derived signals
        self.position = derived_signal_r(self._get_position, x=self.x, y=self.y)
        super().__init__(name=name)

    def _get_position(self, x: float, y: float) -> BeamstopPosition:
        if abs(x) < 1 and abs(y) < 2:
            return BeamstopPosition.IN_POSITION
        else:
            return BeamstopPosition.OUT_OF_POSITION


class MovableBeamstop(Device):
    """As well as reads, this one allows you to move it.

    E.g. bps.mv(beamstop.position, BeamstopPosition.IN_POSITION)
    """

    def __init__(self, name=""):
        # Raw signals
        self.x = soft_signal_rw(float)
        self.y = soft_signal_rw(float)
        # Derived signals
        self.position = derived_signal_rw(
            self._get_position, self._set_from_position, x=self.x, y=self.y
        )
        super().__init__(name=name)

    def _get_position(self, x: float, y: float) -> BeamstopPosition:
        if abs(x) < 1 and abs(y) < 2:
            return BeamstopPosition.IN_POSITION
        else:
            return BeamstopPosition.OUT_OF_POSITION

    async def _set_from_position(self, position: BeamstopPosition) -> None:
        if position == BeamstopPosition.IN_POSITION:
            await asyncio.gather(self.x.set(0), self.y.set(0))
        else:
            await asyncio.gather(self.x.set(3), self.y.set(5))


class Exploder(StandardReadable):
    """This one takes a value and sets all its signal to that value.

    This allows convenience "set all" functions, while the individual
    signals are still free to be set to different values.
    """

    def __init__(self, num_signals: int, name=""):
        with self.add_children_as_readables():
            self.signals = DeviceVector(
                {i: soft_signal_rw(int, units="cts") for i in range(1, num_signals + 1)}
            )
        self.set_all = derived_signal_w(self._set_all, derived_units="cts")
        super().__init__(name=name)

    async def _set_all(self, value: int) -> None:
        coros = [sig.set(value) for sig in self.signals.values()]
        await asyncio.gather(*coros)



from bluesky.utils import plan

from ophyd_async.core import DEFAULT_TIMEOUT, Device, LazyMock, wait_for_connection

from ._wait_for_awaitable import wait_for_awaitable


@plan
def ensure_connected(
    *devices: Device,
    mock: bool | LazyMock = False,
    timeout: float = DEFAULT_TIMEOUT,
    force_reconnect=False,
):
    """Plan stub to ensure devices are connected with a given timeout."""
    device_names = [device.name for device in devices]
    non_unique = {
        device: device.name for device in devices if device_names.count(device.name) > 1
    }
    if non_unique:
        raise ValueError(f"Devices do not have unique names {non_unique}")
    coros = {
        device.name: device.connect(
            mock=mock, timeout=timeout, force_reconnect=force_reconnect
        )
        for device in devices
    }
    yield from wait_for_awaitable(wait_for_connection(**coros))



"""The building blocks for making devices."""

from ._derived_signal import (
    DerivedSignalFactory,
    derived_signal_r,
    derived_signal_rw,
    derived_signal_w,
)
from ._derived_signal_backend import Transform, merge_gathered_dicts
from ._detector import (
    DetectorController,
    DetectorTrigger,
    DetectorWriter,
    StandardDetector,
    TriggerInfo,
)
from ._device import Device, DeviceConnector, DeviceVector, init_devices
from ._device_filler import DeviceFiller
from ._enums import (
    EnabledDisabled,
    EnableDisable,
    InOut,
    OnOff,
    YesNo,
)
from ._flyer import FlyerController, FlyMotorInfo, StandardFlyer
from ._hdf_dataset import HDFDatasetDescription, HDFDocumentComposer
from ._log import config_ophyd_async_logging
from ._mock_signal_backend import MockSignalBackend
from ._protocol import AsyncConfigurable, AsyncReadable, AsyncStageable, Watcher
from ._providers import (
    AutoIncrementFilenameProvider,
    AutoIncrementingPathProvider,
    DatasetDescriber,
    FilenameProvider,
    PathInfo,
    PathProvider,
    StaticFilenameProvider,
    StaticPathProvider,
    UUIDFilenameProvider,
    YMDPathProvider,
)
from ._readable import (
    ConfigSignal,
    HintedSignal,
    StandardReadable,
    StandardReadableFormat,
)
from ._settings import Settings, SettingsProvider
from ._signal import (
    Ignore,
    Signal,
    SignalConnector,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
    observe_signals_value,
    observe_value,
    set_and_wait_for_other_value,
    set_and_wait_for_value,
    soft_signal_r_and_setter,
    soft_signal_rw,
    wait_for_value,
    walk_config_signals,
    walk_devices,
    walk_rw_signals,
    walk_signal_sources,
)
from ._signal_backend import (
    Array1D,
    DTypeScalar_co,
    Primitive,
    SignalBackend,
    SignalDatatype,
    SignalDatatypeT,
    SignalMetadata,
    make_datakey,
)
from ._soft_signal_backend import SoftSignalBackend
from ._status import AsyncStatus, WatchableAsyncStatus, completed_status
from ._table import Table, TableSubclass
from ._utils import (
    CALCULATE_TIMEOUT,
    DEFAULT_TIMEOUT,
    CalculatableTimeout,
    Callback,
    ConfinedModel,
    EnumTypes,
    LazyMock,
    NotConnectedError,
    Reference,
    StrictEnum,
    SubsetEnum,
    SupersetEnum,
    WatcherUpdate,
    error_if_none,
    gather_dict,
    get_dtype,
    get_enum_cls,
    get_unique,
    in_micros,
    wait_for_connection,
)
from ._yaml_settings import YamlSettingsProvider


# Back compat - delete before 1.0
def __getattr__(name):
    import warnings

    renames = {
        "NotConnected": NotConnectedError,
    }
    rename = renames.get(name)
    if rename is not None:
        warnings.warn(
            DeprecationWarning(
                f"{name!r} is deprecated, use {rename.__name__!r} instead"
            ),
            stacklevel=2,
        )
        return rename
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    # Device
    "Device",
    "DeviceConnector",
    "DeviceFiller",
    "DeviceVector",
    "init_devices",
    # Protocols
    "AsyncReadable",
    "AsyncConfigurable",
    "AsyncStageable",
    "Watcher",
    # Status
    "AsyncStatus",
    "WatchableAsyncStatus",
    "WatcherUpdate",
    "completed_status",
    # Signal
    "Signal",
    "SignalR",
    "SignalW",
    "SignalRW",
    "SignalX",
    "SignalBackend",
    "SignalConnector",
    # Signal Types
    "SignalDatatype",
    "SignalDatatypeT",
    "DTypeScalar_co",
    "Array1D",
    "StrictEnum",
    "SubsetEnum",
    "SupersetEnum",
    "EnumTypes",
    "Table",
    "SignalMetadata",
    "Primitive",
    # Soft signal
    "SoftSignalBackend",
    "soft_signal_r_and_setter",
    "soft_signal_rw",
    # Mock signal
    "LazyMock",
    "MockSignalBackend",
    # Signal utilities
    "observe_value",
    "observe_signals_value",
    "wait_for_value",
    "set_and_wait_for_value",
    "set_and_wait_for_other_value",
    "walk_rw_signals",
    "walk_config_signals",
    "walk_devices",
    "walk_signal_sources",
    # Readable
    "StandardReadable",
    "StandardReadableFormat",
    # Detector
    "StandardDetector",
    "TriggerInfo",
    "DetectorTrigger",
    "DetectorController",
    "DetectorWriter",
    # Path
    "PathInfo",
    "PathProvider",
    "StaticPathProvider",
    "AutoIncrementingPathProvider",
    "YMDPathProvider",
    "FilenameProvider",
    "StaticFilenameProvider",
    "AutoIncrementFilenameProvider",
    "UUIDFilenameProvider",
    # Datatset
    "DatasetDescriber",
    "HDFDatasetDescription",
    "HDFDocumentComposer",
    # Flyer
    "StandardFlyer",
    "FlyMotorInfo",
    "FlyerController",
    # Settings
    "Settings",
    "SettingsProvider",
    "YamlSettingsProvider",
    # Utils
    "config_ophyd_async_logging",
    "CALCULATE_TIMEOUT",
    "CalculatableTimeout",
    "DEFAULT_TIMEOUT",
    "Callback",
    "ConfinedModel",
    "NotConnectedError",
    "Reference",
    "error_if_none",
    "gather_dict",
    "get_dtype",
    "get_enum_cls",
    "get_unique",
    "in_micros",
    "make_datakey",
    "wait_for_connection",
    "Ignore",
    # Derived signal
    "derived_signal_r",
    "derived_signal_rw",
    "derived_signal_w",
    "Transform",
    "DerivedSignalFactory",
    "merge_gathered_dicts",
    # Back compat - delete before 1.0
    "ConfigSignal",
    "HintedSignal",
    # Standard enums
    "EnabledDisabled",
    "EnableDisable",
    "InOut",
    "OnOff",
    "YesNo",
    "TableSubclass",
]



from typing import Annotated as A

from ophyd_async.core import SignalR, SignalRW, StrictEnum
from ophyd_async.epics import adcore
from ophyd_async.epics.core import PvSuffix


class PilatusTriggerMode(StrictEnum):
    """Trigger modes for ADPilatus detector."""

    INTERNAL = "Internal"
    EXT_ENABLE = "Ext. Enable"
    EXT_TRIGGER = "Ext. Trigger"
    MULT_TRIGGER = "Mult. Trigger"
    ALIGNMENT = "Alignment"


class PilatusDriverIO(adcore.ADBaseIO):
    """Driver for the Pilatus pixel array detectors."""

    """This mirrors the interface provided by ADPilatus/db/pilatus.template."""
    """See HTML docs at https://areadetector.github.io/areaDetector/ADPilatus/pilatusDoc.html"""
    trigger_mode: A[SignalRW[PilatusTriggerMode], PvSuffix.rbv("TriggerMode")]
    armed: A[SignalR[bool], PvSuffix("Armed")]



from ophyd_async.core import StandardReadable

from ._motor import DemoMotor


class DemoStage(StandardReadable):
    """A simulated sample stage with X and Y movables."""

    def __init__(self, prefix: str, name="") -> None:
        # Define some child Devices
        with self.add_children_as_readables():
            self.x = DemoMotor(prefix + "X:")
            self.y = DemoMotor(prefix + "Y:")
        # Set name of device and child devices
        super().__init__(name=name)



"""Equivalent of bluesky.protocols.Status for asynchronous tasks."""

from __future__ import annotations

import asyncio
import functools
import time
from collections.abc import AsyncIterator, Awaitable, Callable, Coroutine
from dataclasses import asdict, replace
from typing import Generic

from bluesky.protocols import Status

from ._device import Device
from ._protocol import Watcher
from ._utils import Callback, P, T, WatcherUpdate


class AsyncStatusBase(Status, Awaitable[None]):
    """Convert asyncio awaitable to bluesky Status interface."""

    def __init__(self, awaitable: Coroutine | asyncio.Task, name: str | None = None):
        if isinstance(awaitable, asyncio.Task):
            self.task = awaitable
        else:
            self.task = asyncio.create_task(awaitable)
        self.task.add_done_callback(self._run_callbacks)
        self._callbacks: list[Callback[Status]] = []
        self._name = name

    def __await__(self):
        return self.task.__await__()

    def add_callback(self, callback: Callback[Status]):
        if self.done:
            callback(self)
        else:
            self._callbacks.append(callback)

    def _run_callbacks(self, task: asyncio.Task):
        for callback in self._callbacks:
            callback(self)

    def exception(self, timeout: float | None = 0.0) -> BaseException | None:
        """Return any exception raised by the task.

        :param timeout:
            Taken for compatibility with the Status interface, but must be 0.0 as we
            cannot wait for an async function in a sync call.
        """
        if timeout != 0.0:
            raise ValueError(
                "cannot honour any timeout other than 0 in an asynchronous function"
            )
        if self.task.done():
            try:
                return self.task.exception()
            except asyncio.CancelledError as exc:
                return exc
        return None

    @property
    def done(self) -> bool:
        return self.task.done()

    @property
    def success(self) -> bool:
        return (
            self.task.done()
            and not self.task.cancelled()
            and self.task.exception() is None
        )

    def __repr__(self) -> str:
        if self.done:
            if e := self.exception():
                status = f"errored: {repr(e)}"
            else:
                status = "done"
        else:
            status = "pending"
        device_str = f"device: {self._name}, " if self._name else ""
        return (
            f"<{type(self).__name__}, {device_str}"
            f"task: {self.task.get_coro()}, {status}>"
        )

    __str__ = __repr__


class AsyncStatus(AsyncStatusBase):
    """Convert an asyncio awaitable to bluesky Status interface.

    :param awaitable: The coroutine or task to await.
    :param name: The name of the device, if available.

    For example:
    ```python
    status = AsyncStatus(asyncio.sleep(1))
    assert not status.done
    await status # waits for 1 second
    assert status.done
    ```
    """

    @classmethod
    def wrap(cls, f: Callable[P, Coroutine]) -> Callable[P, AsyncStatus]:
        """Wrap an async function in an AsyncStatus and return it.

        Used to make an async function conform to a bluesky protocol.

        For example:
        ```python
        class MyDevice(Device):
            @AsyncStatus.wrap
            async def trigger(self):
                await asyncio.sleep(1)
        ```
        """

        @functools.wraps(f)
        def wrap_f(*args: P.args, **kwargs: P.kwargs) -> AsyncStatus:
            if args and isinstance(args[0], Device):
                name = args[0].name
            else:
                name = None
            return cls(f(*args, **kwargs), name=name)

        return wrap_f


class WatchableAsyncStatus(AsyncStatusBase, Generic[T]):
    """Convert an asyncio async iterable to bluesky Status and Watcher interface.

    :param iterator: The async iterable to await.
    :param name: The name of the device, if available.
    """

    def __init__(
        self, iterator: AsyncIterator[WatcherUpdate[T]], name: str | None = None
    ):
        self._watchers: list[Watcher] = []
        self._start = time.monotonic()
        self._last_update: WatcherUpdate[T] | None = None
        super().__init__(self._notify_watchers_from(iterator), name)

    async def _notify_watchers_from(self, iterator: AsyncIterator[WatcherUpdate[T]]):
        async for update in iterator:
            self._last_update = (
                update
                if update.time_elapsed is not None
                else replace(update, time_elapsed=time.monotonic() - self._start)
            )
            for watcher in self._watchers:
                self._update_watcher(watcher, self._last_update)

    def _update_watcher(self, watcher: Watcher, update: WatcherUpdate[T]):
        vals = asdict(
            update, dict_factory=lambda d: {k: v for k, v in d if v is not None}
        )
        watcher(**vals)

    def watch(self, watcher: Watcher):
        """Add a watcher to the status.

        It is called:
        - immediately if there has already been an update
        - on every subsequent update
        """
        self._watchers.append(watcher)
        if self._last_update:
            self._update_watcher(watcher, self._last_update)

    @classmethod
    def wrap(
        cls,
        f: Callable[P, AsyncIterator[WatcherUpdate[T]]],
    ) -> Callable[P, WatchableAsyncStatus[T]]:
        """Wrap an AsyncIterator in a WatchableAsyncStatus.

        For example:
        ```python
        class MyDevice(Device):
            @WatchableAsyncStatus.wrap
            async def trigger(self):
                # sleep for a second, updating on progress every 0.1 seconds
                for i in range(10):
                    yield WatcherUpdate(initial=0, current=i*0.1, target=1)
                    await asyncio.sleep(0.1)
        ```
        """

        @functools.wraps(f)
        def wrap_f(*args: P.args, **kwargs: P.kwargs) -> WatchableAsyncStatus[T]:
            if args and isinstance(args[0], Device):
                name = args[0].name
            else:
                name = None
            return cls(f(*args, **kwargs), name=name)

        return wrap_f


@AsyncStatus.wrap
async def completed_status(exception: Exception | None = None):
    """Return a completed AsyncStatus.

    :param exception: If given, then raise this exception when awaited.
    """
    if exception:
        raise exception
    return None



import pytest

from ophyd_async.core import DetectorTrigger, TriggerInfo, init_devices
from ophyd_async.epics import adcore
from ophyd_async.testing import set_mock_value


@pytest.fixture
def cont_acq_controller(RE):
    with init_devices(mock=True):
        drv = adcore.ADBaseIO("DRV")
        cb_plugin = adcore.NDPluginCBIO("CB1")

    set_mock_value(drv.image_mode, adcore.ADImageMode.CONTINUOUS)
    set_mock_value(drv.acquire_time, 0.8)
    set_mock_value(drv.acquire_period, 1.0)
    set_mock_value(drv.acquire, True)
    return adcore.ADBaseContAcqController(drv, cb_plugin)


def generate_one_shot_trigger_info(trigger_mode=DetectorTrigger.INTERNAL, livetime=0.8):
    if trigger_mode != DetectorTrigger.INTERNAL:
        return TriggerInfo(
            number_of_events=1,
            trigger=trigger_mode,
            livetime=livetime,
            deadtime=0.001,
        )
    else:
        return TriggerInfo(number_of_events=1, trigger=trigger_mode, livetime=livetime)


async def test_cont_acq_controller_invalid_trigger_mode(
    cont_acq_controller: adcore.ADBaseContAcqController,
):
    trigger_info = generate_one_shot_trigger_info(
        trigger_mode=DetectorTrigger.CONSTANT_GATE
    )
    with pytest.raises(TypeError) as exc:
        await cont_acq_controller.prepare(trigger_info)
    assert (
        str(exc.value)
        == "The continuous acq interface only supports internal triggering."
    )


async def test_cont_acq_controller_invalid_exp_time(
    cont_acq_controller: adcore.ADBaseContAcqController,
):
    with pytest.raises(ValueError) as exc:
        await cont_acq_controller.prepare(generate_one_shot_trigger_info(livetime=0.1))
    assert (
        str(exc.value)
        == "Detector exposure time currently set to 0.8, but requested exposure is 0.1"
    )


async def test_cont_acq_controller_not_in_continuous_mode(
    cont_acq_controller: adcore.ADBaseContAcqController,
):
    set_mock_value(cont_acq_controller.driver.image_mode, adcore.ADImageMode.SINGLE)

    with pytest.raises(RuntimeError) as exc:
        await cont_acq_controller.prepare(generate_one_shot_trigger_info())
    assert (
        str(exc.value)
        == "Driver must be acquiring in continuous mode to use the cont acq interface"
    )


async def test_cont_acq_controller_not_acquiring(
    cont_acq_controller: adcore.ADBaseContAcqController,
):
    set_mock_value(cont_acq_controller.driver.acquire, False)

    with pytest.raises(RuntimeError) as exc:
        await cont_acq_controller.prepare(generate_one_shot_trigger_info())
    assert (
        str(exc.value)
        == "Driver must be acquiring in continuous mode to use the cont acq interface"
    )



import asyncio

from ophyd_async.core import DetectorTrigger, OnOff, TriggerInfo
from ophyd_async.epics import adcore

from ._aravis_io import AravisDriverIO, AravisTriggerSource

# The deadtime of an ADaravis controller varies depending on the exact model of camera.
# Ideally we would maximize performance by dynamically retrieving the deadtime at
# runtime. See https://github.com/bluesky/ophyd-async/issues/308
_HIGHEST_POSSIBLE_DEADTIME = 1961e-6


class AravisController(adcore.ADBaseController[AravisDriverIO]):
    """`DetectorController` for an `AravisDriverIO`."""

    def get_deadtime(self, exposure: float | None) -> float:
        return _HIGHEST_POSSIBLE_DEADTIME

    async def prepare(self, trigger_info: TriggerInfo) -> None:
        if (exposure := trigger_info.livetime) is not None:
            await self.driver.acquire_time.set(exposure)

        if trigger_info.trigger is DetectorTrigger.INTERNAL:
            # Set trigger mode off to ignore the trigger source
            await self.driver.trigger_mode.set(OnOff.OFF)
        elif trigger_info.trigger in {
            DetectorTrigger.CONSTANT_GATE,
            DetectorTrigger.EDGE_TRIGGER,
        }:
            # Trigger on the rising edge of Line1
            # trigger mode must be set first and on it's own!
            await self.driver.trigger_mode.set(OnOff.ON)
            await self.driver.trigger_source.set(AravisTriggerSource.LINE1)
        else:
            raise ValueError(f"ADAravis does not support {trigger_info.trigger}")

        if trigger_info.total_number_of_exposures == 0:
            image_mode = adcore.ADImageMode.CONTINUOUS
        else:
            image_mode = adcore.ADImageMode.MULTIPLE
        await asyncio.gather(
            self.driver.num_images.set(trigger_info.total_number_of_exposures),
            self.driver.image_mode.set(image_mode),
        )



from __future__ import annotations

from typing import Literal, cast

from ophyd_async.core import (
    Device,
    DeviceConnector,
    DeviceFiller,
    LazyMock,
    Signal,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
)

from ._epics_connector import fill_backend_with_prefix
from ._signal import PvaSignalBackend, pvget_with_timeout

Entry = dict[str, str]

OldPVIVector = list[Entry | None]
# The older PVI structure has vectors of the form
# structure[] ttlout
#     (none)
#     structure
#         string d PANDABLOCKS_IOC:TTLOUT1:PVI
#     structure
#         string d PANDABLOCKS_IOC:TTLOUT2:PVI
#     structure
#         string d PANDABLOCKS_IOC:TTLOUT3:PVI


FastCSPVIVector = dict[Literal["d"], Entry]
# The newer pva FastCS PVI structure has vectors of the form
# structure ttlout
#     structure d
#         string v1 FASTCS_PANDA:Ttlout1:PVI
#         string v2 FASTCS_PANDA:Ttlout2:PVI
#         string v3 FASTCS_PANDA:Ttlout3:PVI
#         string v4 FASTCS_PANDA:Ttlout4:PVI


def _get_signal_details(entry: Entry) -> tuple[type[Signal], str, str]:
    match entry:
        case {"r": read_pv, "w": write_pv}:
            return SignalRW, read_pv, write_pv
        case {"r": read_pv}:
            return SignalR, read_pv, read_pv
        case {"w": write_pv}:
            return SignalW, write_pv, write_pv
        case {"rw": read_write_pv}:
            return SignalRW, read_write_pv, read_write_pv
        case {"x": execute_pv}:
            return SignalX, execute_pv, execute_pv
        case _:
            raise TypeError(f"Can't process entry {entry}")


def _is_device_vector_entry(entry: Entry | OldPVIVector | FastCSPVIVector) -> bool:
    return isinstance(entry, list) or (
        entry.keys() == {"d"} and isinstance(entry["d"], dict)
    )


class PviDeviceConnector(DeviceConnector):
    """Connect to PVI structure served over PVA.

    At init, fill in all the type hinted signals. At connection check their
    types and fill in any extra signals.

    :param prefix:
        The PV prefix of the device, "PVI" will be appended to it to get the PVI
        PV.
    :param error_hint:
        If given, this will be appended to the error message if any of they type
        hinted Signals are not present.
    """

    mock_device_vector_len: int = 2

    def __init__(self, prefix: str = "", error_hint: str = "") -> None:
        # TODO: what happens if we get a leading "pva://" here?
        self.prefix = prefix
        self.pvi_pv = prefix + "PVI"
        self.error_hint = error_hint

    def create_children_from_annotations(self, device: Device):
        if not hasattr(self, "filler"):
            self.filler = DeviceFiller(
                device=device,
                signal_backend_factory=PvaSignalBackend,
                device_connector_factory=PviDeviceConnector,
            )
            # Devices will be created with unfilled PviDeviceConnectors
            list(self.filler.create_devices_from_annotations(filled=False))
            # Signals can be filled in with EpicsSignalSuffix and checked at runtime
            for backend, annotations in self.filler.create_signals_from_annotations(
                filled=False
            ):
                fill_backend_with_prefix(self.prefix, backend, annotations)
            self.filler.check_created()

    def _fill_child(self, name: str, entry: Entry, vector_index: int | None = None):
        if set(entry) == {"d"}:
            connector = self.filler.fill_child_device(name, vector_index=vector_index)
            connector.pvi_pv = entry["d"]
        else:
            signal_type, read_pv, write_pv = _get_signal_details(entry)
            backend = self.filler.fill_child_signal(name, signal_type, vector_index)
            backend.read_pv = read_pv
            backend.write_pv = write_pv

    async def connect_mock(self, device: Device, mock: LazyMock):
        self.filler.create_device_vector_entries_to_mock(self.mock_device_vector_len)
        # Set the name of the device to name all children
        device.set_name(device.name)
        return await super().connect_mock(device, mock)

    def _fill_vector_child(self, name: str, entry: OldPVIVector | FastCSPVIVector):
        if isinstance(entry, list):
            for i, e in enumerate(entry):
                if e:
                    self._fill_child(name, e, i)
        else:
            for i_string, e in entry["d"].items():
                self._fill_child(name, {"d": e}, int(i_string.lstrip("v")))

    async def connect_real(
        self, device: Device, timeout: float, force_reconnect: bool
    ) -> None:
        pvi_structure = await pvget_with_timeout(self.pvi_pv, timeout)

        entries: dict[str, Entry | OldPVIVector | FastCSPVIVector] = pvi_structure[
            "value"
        ].todict()
        # Fill based on what PVI gives us
        for name, entry in entries.items():
            if _is_device_vector_entry(entry):
                self._fill_vector_child(
                    name, cast(OldPVIVector | FastCSPVIVector, entry)
                )
            else:
                # This is a child
                self._fill_child(name, cast(Entry, entry))

        # Check that all the requested children have been filled
        suffix = f"\n{self.error_hint}" if self.error_hint else ""
        self.filler.check_filled(f"{self.pvi_pv}: {entries}{suffix}")
        # Set the name of the device to name all children
        device.set_name(device.name)
        return await super().connect_real(device, timeout, force_reconnect)



from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import Annotated, Any, TypeVar, get_origin, get_type_hints

import numpy as np
from pydantic import ConfigDict, Field, model_validator
from pydantic_numpy.helper.annotation import NpArrayPydanticAnnotation

from ._utils import ConfinedModel, get_dtype

TableSubclass = TypeVar("TableSubclass", bound="Table")


def _concat(value1, value2):
    """Concatenate two values, supports both NumPy arrays and generic types."""
    if isinstance(value1, np.ndarray):
        return np.concatenate((value1, value2))
    else:
        return value1 + value2


def _make_default_factory(dtype: np.dtype) -> Callable[[], np.ndarray]:
    """Creates a default factory, returns an empty Numpy array of a specified dtype."""

    def numpy_array_default_factory() -> np.ndarray:
        return np.array([], dtype)

    return numpy_array_default_factory


class Table(ConfinedModel):
    """An abstraction of a Table where each field is a column.

    For example:
    ```python
    >>> from ophyd_async.core import Table, Array1D
    >>> import numpy as np
    >>> from collections.abc import Sequence
    >>> class MyTable(Table):
    ...     a: Array1D[np.int8]
    ...     b: Sequence[str]
    ...
    >>> t = MyTable(a=[1, 2], b=["x", "y"])
    >>> len(t)  # the length is the number of rows
    2
    >>> t2 = t + t  # adding tables together concatenates them
    >>> t2.a
    array([1, 2, 1, 2], dtype=int8)
    >>> t2.b
    ['x', 'y', 'x', 'y']
    >>> t2[1]  # slice a row
    array([(2, b'y')], dtype=[('a', 'i1'), ('b', 'S40')])

    ```
    """

    # You can use Table in 2 ways:
    # 1. Table(**whatever_pva_gives_us) when pvi adds a Signal to a Device that is not
    #    type hinted
    # 2. MyTable(**whatever_pva_gives_us) where the Signal is type hinted
    #
    # For 1 we want extra="allow" so it is passed through as is. There are no base class
    # fields, only "extra" fields, so they must be allowed. For 2 we want extra="forbid"
    # so it is strictly checked against the BaseModel we are supplied.
    model_config = ConfigDict(extra="allow")

    # Add an init method to match the above model config, otherwise the type
    # checker will not think we can pass arbitrary kwargs into the base class init...
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def __init_subclass__(cls):
        # ...but forbid extra in subclasses so it gets validated
        cls.model_config = ConfigDict(validate_assignment=True, extra="forbid")
        # Change fields to have the correct annotations
        # TODO: refactor so we don't need this to break circular imports
        from ._signal_backend import Array1D

        for k, anno in get_type_hints(cls, localns={"Array1D": Array1D}).items():
            if get_origin(anno) is np.ndarray:
                dtype = get_dtype(anno)
                new_anno = Annotated[
                    anno,
                    NpArrayPydanticAnnotation.factory(
                        data_type=dtype.type, dimensions=1, strict_data_typing=False
                    ),
                    Field(default_factory=_make_default_factory(dtype)),
                ]
            elif get_origin(anno) is Sequence:
                new_anno = Annotated[anno, Field(default_factory=list)]
            else:
                raise TypeError(f"Cannot use annotation {anno} in a Table")
            cls.__annotations__[k] = new_anno

    @classmethod
    def empty(cls: type[TableSubclass]) -> TableSubclass:
        """Makes an empty table with zero length columns."""
        return cls()  # type: ignore

    def __add__(self, right: TableSubclass) -> TableSubclass:
        """Concatenate the arrays in field values."""
        cls = type(right)
        if type(self) is not cls:
            raise RuntimeError(
                f"{right} is not a `Table`, or is not the same "
                f"type of `Table` as {self}."
            )

        return cls(
            **{
                field_name: _concat(
                    getattr(self, field_name), getattr(right, field_name)
                )
                for field_name in cls.model_fields
            }
        )

    def numpy_dtype(self) -> np.dtype:
        """Return a numpy dtype for a single row."""
        dtype = []
        for k, v in self:
            if isinstance(v, np.ndarray):
                dtype.append((k, v.dtype))
            else:
                # TODO: use np.dtypes.StringDType when we can use in structured arrays
                # https://github.com/numpy/numpy/issues/25693
                dtype.append((k, np.dtype("S40")))
        return np.dtype(dtype)

    def numpy_table(self, selection: slice | None = None) -> np.ndarray:
        """Return a numpy array of the whole table."""
        array = None
        for k, v in self:
            if selection:
                v = v[selection]
            if array is None:
                array = np.empty(v.shape, dtype=self.numpy_dtype())
            array[k] = v  # type: ignore
        if array is None:
            msg = "No arrays found in table"
            raise ValueError(msg)
        return array

    @model_validator(mode="before")
    @classmethod
    def _validate_array_dtypes(cls, data: Any) -> Any:
        # Validates that array datatypes given in the table are of the
        # correct format.
        if isinstance(data, dict):
            data_dict = data
        elif isinstance(data, Table):
            data_dict = data.model_dump()
        else:
            raise AssertionError(f"Cannot construct Table from {data}")
        for field_name, field_value in cls.model_fields.items():
            if (
                get_origin(field_value.annotation) is np.ndarray
                and field_value.annotation
                and field_name in data_dict
            ):
                data_value = data_dict[field_name]
                expected_dtype = get_dtype(field_value.annotation)
                # Convert to correct dtype, but only if we don't lose precision
                # as a result
                cast_value = np.array(data_value).astype(expected_dtype)
                if not np.array_equal(data_value, cast_value):
                    msg = (
                        f"{field_name}: Cannot cast {data_value} to {expected_dtype} "
                        "without losing precision"
                    )
                    raise ValueError(msg)
                data_dict[field_name] = cast_value
        return data_dict

    @model_validator(mode="after")
    def _validate_lengths(self) -> Table:
        lengths: dict[int, set[str]] = {}
        for field_name, field_value in self:
            lengths.setdefault(len(field_value), set()).add(field_name)
        if len(lengths) > 1:
            msg = f"Columns should be same length, got {lengths=}"
            raise ValueError(msg)
        return self

    def __len__(self) -> int:
        return len(next(iter(self))[1])

    def __getitem__(self, item: int | slice) -> np.ndarray:
        if isinstance(item, int):
            return self.numpy_table(slice(item, item + 1))
        else:
            return self.numpy_table(item)



"""Support for the ADAravis areaDetector driver.

https://github.com/areaDetector/ADSimDetector
"""

from ._sim import SimDetector
from ._sim_controller import SimController
from ._sim_io import SimDriverIO

__all__ = [
    "SimDriverIO",
    "SimController",
    "SimDetector",
]



import asyncio
from typing import TypedDict

from bluesky.protocols import Movable

from ophyd_async.core import AsyncStatus, DerivedSignalFactory, Device

from ._mirror_vertical import TwoJackDerived, TwoJackTransform
from ._motor import SimMotor


class HorizontalMirrorDerived(TypedDict):
    x: float
    roll: float


class HorizontalMirror(Device, Movable):
    def __init__(self, name=""):
        # Raw signals
        self.x1 = SimMotor()
        self.x2 = SimMotor()
        # Parameter
        # This could also be set as 'soft_signal_rw(float, initial_value=1)'
        self.x1_x2_distance = 1.0
        # Derived signals
        self._factory = DerivedSignalFactory(
            TwoJackTransform,
            self._set_mirror,
            jack1=self.x1,
            jack2=self.x2,
            distance=self.x1_x2_distance,
        )
        self.x = self._factory.derived_signal_rw(float, "height")
        self.roll = self._factory.derived_signal_rw(float, "angle")
        super().__init__(name=name)

    async def _set_mirror(self, derived: TwoJackDerived) -> None:
        transform = await self._factory.transform()
        raw = transform.derived_to_raw(**derived)
        await asyncio.gather(
            self.x1.set(raw["jack1"]),
            self.x2.set(raw["jack2"]),
        )

    @AsyncStatus.wrap
    async def set(self, value: HorizontalMirrorDerived) -> None:
        await self._set_mirror(TwoJackDerived(height=value["x"], angle=value["roll"]))



from typing import cast

import pytest

from ophyd_async.core import DetectorTrigger, OnOff, TriggerInfo
from ophyd_async.epics import adcore, advimba
from ophyd_async.epics.advimba import (
    VimbaExposeOutMode,
    VimbaTriggerSource,
)
from ophyd_async.testing import set_mock_value


@pytest.fixture
def test_advimba(ad_standard_det_factory) -> advimba.VimbaDetector:
    return ad_standard_det_factory(advimba.VimbaDetector, adcore.ADHDFWriter)


async def test_get_deadtime(
    test_advimba: advimba.VimbaDetector,
):
    # Currently Vimba controller doesn't support getting deadtime.
    assert test_advimba._controller.get_deadtime(0) == 0.001


async def test_arming_trig_modes(test_advimba: advimba.VimbaDetector):
    driver = cast(advimba.VimbaDriverIO, test_advimba.driver)

    set_mock_value(driver.trigger_source, VimbaTriggerSource.FREERUN)
    set_mock_value(driver.trigger_mode, OnOff.OFF)
    set_mock_value(driver.exposure_mode, VimbaExposeOutMode.TIMED)

    async def setup_trigger_mode(trig_mode: DetectorTrigger):
        await test_advimba._controller.prepare(
            TriggerInfo(number_of_events=1, trigger=trig_mode)
        )
        await test_advimba._controller.arm()
        await test_advimba._controller.wait_for_idle()
        # Prevent timeouts
        set_mock_value(driver.acquire, True)

    # Default TriggerSource
    assert (await driver.trigger_source.get_value()) == VimbaTriggerSource.FREERUN
    assert (await driver.trigger_mode.get_value()) == OnOff.OFF
    assert (await driver.exposure_mode.get_value()) == VimbaExposeOutMode.TIMED

    await setup_trigger_mode(DetectorTrigger.EDGE_TRIGGER)
    assert (await driver.trigger_source.get_value()) == VimbaTriggerSource.LINE1
    assert (await driver.trigger_mode.get_value()) == OnOff.ON
    assert (await driver.exposure_mode.get_value()) == VimbaExposeOutMode.TIMED

    await setup_trigger_mode(DetectorTrigger.CONSTANT_GATE)
    assert (await driver.trigger_source.get_value()) == VimbaTriggerSource.LINE1
    assert (await driver.trigger_mode.get_value()) == OnOff.ON
    assert (await driver.exposure_mode.get_value()) == VimbaExposeOutMode.TRIGGER_WIDTH

    await setup_trigger_mode(DetectorTrigger.INTERNAL)
    assert (await driver.trigger_source.get_value()) == VimbaTriggerSource.FREERUN
    assert (await driver.trigger_mode.get_value()) == OnOff.OFF
    assert (await driver.exposure_mode.get_value()) == VimbaExposeOutMode.TIMED

    await setup_trigger_mode(DetectorTrigger.VARIABLE_GATE)
    assert (await driver.trigger_source.get_value()) == VimbaTriggerSource.LINE1
    assert (await driver.trigger_mode.get_value()) == OnOff.ON
    assert (await driver.exposure_mode.get_value()) == VimbaExposeOutMode.TRIGGER_WIDTH



from ._kinetix import KinetixDetector
from ._kinetix_controller import KinetixController
from ._kinetix_io import KinetixDriverIO, KinetixReadoutMode, KinetixTriggerMode

__all__ = [
    "KinetixDetector",
    "KinetixController",
    "KinetixDriverIO",
    "KinetixTriggerMode",
    "KinetixReadoutMode",
]



import asyncio

from ophyd_async.core import (
    AsyncStatus,
    DeviceVector,
    StandardReadable,
)

from ._counter import TangoCounter
from ._mover import TangoMover


class TangoDetector(StandardReadable):
    """For use with tango detector devices."""

    def __init__(self, mover_trl: str, counter_trls: list[str], name=""):
        # A detector device may be composed of tango sub-devices
        self.mover = TangoMover(mover_trl)
        self.counters = DeviceVector(
            {i + 1: TangoCounter(c_trl) for i, c_trl in enumerate(counter_trls)}
        )

        # Define the readables for TangoDetector
        # DeviceVectors are incompatible with AsyncReadable. Ignore until fixed.
        self.add_readables([self.counters, self.mover])  # type: ignore

        super().__init__(name=name)

    def set(self, value):
        return self.mover.set(value)

    def stop(self, success: bool = True) -> AsyncStatus:
        return self.mover.stop(success)

    @AsyncStatus.wrap
    async def trigger(self):
        statuses = []
        for counter in self.counters.values():
            statuses.append(counter.reset())
        await asyncio.gather(*statuses)
        statuses.clear()
        for counter in self.counters.values():
            statuses.append(counter.trigger())
        await asyncio.gather(*statuses)



from collections.abc import Sequence

from ophyd_async.core import PathProvider, SignalR, StandardDetector

from ._blob_detector_controller import BlobDetectorController
from ._blob_detector_writer import BlobDetectorWriter
from ._pattern_generator import PatternGenerator


class SimBlobDetector(StandardDetector):
    """Simulates a detector and writes Blobs to file."""

    def __init__(
        self,
        path_provider: PathProvider,
        pattern_generator: PatternGenerator | None = None,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
    ) -> None:
        self.pattern_generator = pattern_generator or PatternGenerator()

        super().__init__(
            controller=BlobDetectorController(
                pattern_generator=self.pattern_generator,
            ),
            writer=BlobDetectorWriter(
                pattern_generator=self.pattern_generator,
                path_provider=path_provider,
            ),
            config_sigs=config_sigs,
            name=name,
        )



from collections.abc import Iterator

from bluesky.protocols import StreamAsset
from event_model import (  # type: ignore
    ComposeStreamResource,
    ComposeStreamResourceBundle,
    StreamRange,
)
from pydantic import Field

from ._utils import ConfinedModel


class HDFDatasetDescription(ConfinedModel):
    """A description of the type and shape of a dataset in an HDF file."""

    data_key: str
    """The data_key that will appear in the event descriptor,
    e.g. det or det.data"""

    dataset: str
    """The dataset name within the HDF file,
    e.g. /entry/data/data or /entry/instrument/NDAttributes/sum"""

    shape: tuple[int | None, ...] = Field(default_factory=tuple)
    """The shape of a single event's data in the HDF file,
    e.g. (1, 768, 1024) for arrays or () for scalars"""

    dtype_numpy: str = ""
    """The numpy dtype for this field,
    e.g. <i2 or <f8"""

    chunk_shape: tuple[int, ...]
    """The explicit chunk size written to disk"""


SLICE_NAME = "AD_HDF5_SWMR_SLICE"


class HDFDocumentComposer:
    """A helper class to make stream resource and datums for HDF datasets.

    :param full_file_name: Absolute path to the file that has been written
    :param datasets: Descriptions of each of the datasets that will appear in the file
    """

    def __init__(
        self,
        file_uri: str,
        datasets: list[HDFDatasetDescription],
    ) -> None:
        self._last_emitted = 0
        uri = file_uri
        bundler_composer = ComposeStreamResource()
        self._bundles: list[ComposeStreamResourceBundle] = [
            bundler_composer(
                mimetype="application/x-hdf5",
                uri=uri,
                data_key=ds.data_key,
                parameters={
                    "dataset": ds.dataset,
                    "chunk_shape": ds.chunk_shape,
                },
                uid=None,
                validate=True,
            )
            for ds in datasets
        ]

    def make_stream_docs(self, indices_written: int) -> Iterator[StreamAsset]:
        # TODO: fail if we get dropped frames
        if indices_written and not self._last_emitted:
            for bundle in self._bundles:
                yield "stream_resource", bundle.stream_resource_doc

        if indices_written > self._last_emitted:
            indices: StreamRange = {
                "start": self._last_emitted,
                "stop": indices_written,
            }
            self._last_emitted = indices_written
            for bundle in self._bundles:
                yield "stream_datum", bundle.compose_stream_datum(indices)



import atexit
from pathlib import Path

from ophyd_async.epics.testing import TestingIOC

HERE = Path(__file__).absolute().parent


def start_ioc_subprocess(prefix: str, num_channels: int) -> TestingIOC:
    """Start an IOC subprocess for sample stage and sensor.

    :param prefix: The prefix for the IOC PVs.
    :param num_channels: The number of point detector channels to create.
    """
    ioc = TestingIOC()
    # Create X and Y motors
    for suffix in ["X", "Y"]:
        ioc.add_database(HERE / "motor.db", P=f"{prefix}STAGE:{suffix}:")
    # Create a multichannel counter with num_counters
    ioc.add_database(HERE / "point_detector.db", P=f"{prefix}DET:")
    for i in range(1, num_channels + 1):
        ioc.add_database(
            HERE / "point_detector_channel.db",
            P=f"{prefix}DET:",
            CHANNEL=str(i),
            X=f"{prefix}STAGE:X:",
            Y=f"{prefix}STAGE:Y:",
        )
    # Start IOC and register it to be stopped at exit
    ioc.start()
    atexit.register(ioc.stop)
    return ioc



import textwrap
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

import numpy as np
from tango import AttrDataFormat, AttrWriteType, DevState
from tango.server import Device, attribute, command

from ophyd_async.core import (
    Array1D,
    DTypeScalar_co,
    StrictEnum,
)
from ophyd_async.testing import float_array_value, int_array_value

T = TypeVar("T")


class ExampleStrEnum(StrictEnum):
    A = "AAA"
    B = "BBB"
    C = "CCC"


def int_image_value(
    dtype: type[DTypeScalar_co],
):
    # how do we type this?
    array_1d = int_array_value(dtype)
    return np.vstack((array_1d, array_1d))


def float_image_value(
    dtype: type[DTypeScalar_co],
):
    # how do we type this?
    array_1d = float_array_value(dtype)
    return np.vstack((array_1d, array_1d))


def _valid_command(dformat: AttrDataFormat, dtype: str):
    if dtype == "DevUChar":
        return False
    if dformat != AttrDataFormat.SCALAR and dtype in ["DevState", "DevEnum"]:
        return False
    return True


@dataclass
class AttributeData(Generic[T]):
    name: str
    tango_type: str
    initial_scalar: T
    initial_spectrum: Array1D | Sequence[T]


_all_attribute_definitions = [
    AttributeData(
        "str",
        "DevString",
        "test_string",
        ["one", "two", "three"],
    ),
    AttributeData(
        "bool",
        "DevBoolean",
        True,
        np.array([False, True], dtype=bool),
    ),
    AttributeData("strenum", "DevEnum", 1, np.array([0, 1, 2])),
    AttributeData("int8", "DevShort", 1, int_array_value(np.int8)),
    AttributeData("uint8", "DevUChar", 1, int_array_value(np.uint8)),
    AttributeData("int16", "DevShort", 1, int_array_value(np.int16)),
    AttributeData("uint16", "DevUShort", 1, int_array_value(np.uint16)),
    AttributeData("int32", "DevLong", 1, int_array_value(np.int32)),
    AttributeData("uint32", "DevULong", 1, int_array_value(np.uint32)),
    AttributeData("int64", "DevLong64", 1, int_array_value(np.int64)),
    AttributeData("uint64", "DevULong64", 1, int_array_value(np.uint64)),
    AttributeData("float32", "DevFloat", 1.234, float_array_value(np.float32)),
    AttributeData("float64", "DevDouble", 1.234, float_array_value(np.float64)),
    AttributeData(
        "my_state",
        "DevState",
        DevState.INIT,
        np.array(list(DevState.names.values()), dtype=DevState),
    ),
]


class OneOfEverythingTangoDevice(Device):
    attr_values = {}
    initial_values = {}

    def _add_attr(self, attr: attribute, initial_value):
        self.attr_values[attr.name] = initial_value
        self.initial_values[attr.name] = initial_value
        self.add_attribute(attr)
        self.set_change_event(attr.name, True, False)

    def add_scalar_attr(self, name: str, dtype: str, initial_value: Any):
        attr = attribute(
            name=name,
            dtype=dtype,
            dformat=AttrDataFormat.SCALAR,
            access=AttrWriteType.READ_WRITE,
            fget=self.read,
            fset=self.write,
            enum_labels=[e.value for e in ExampleStrEnum],
        )
        self._add_attr(attr, initial_value)

    def add_array_attrs(self, name: str, dtype: str, initial_value: np.ndarray):
        spectrum_name = f"{name}_spectrum"
        if hasattr(initial_value, "shape"):
            max_dim_x = initial_value.shape[-1]
        else:
            max_dim_x = len(initial_value)
        spectrum_attr = attribute(
            name=spectrum_name,
            dtype=dtype,
            dformat=AttrDataFormat.SPECTRUM,
            access=AttrWriteType.READ_WRITE,
            fget=self.read,
            fset=self.write,
            max_dim_x=max_dim_x,
            enum_labels=[e.value for e in ExampleStrEnum],
        )
        image_name = f"{name}_image"
        image_attr = attribute(
            name=image_name,
            dtype=dtype,
            dformat=AttrDataFormat.IMAGE,
            access=AttrWriteType.READ_WRITE,
            fget=self.read,
            fset=self.write,
            max_dim_x=max_dim_x,
            max_dim_y=2,
            enum_labels=[e.value for e in ExampleStrEnum],
        )
        self._add_attr(spectrum_attr, initial_value)
        # have image just be 2 of the initial spectrum stacked
        # String images are not supported, do not add their attribute data
        if name in ["str", "strenum", "my_state"]:
            return
        self._add_attr(image_attr, np.vstack((initial_value, initial_value)))

    def add_scalar_command(self, name: str, dtype: str):
        if _valid_command(AttrDataFormat.SCALAR, dtype):
            self.add_command(
                command(
                    f=getattr(self, f"{name}_cmd"),
                    dtype_in=dtype,
                    dtype_out=dtype,
                    dformat_in=AttrDataFormat.SCALAR,
                    dformat_out=AttrDataFormat.SCALAR,
                ),
            )

    def add_spectrum_command(self, name: str, dtype: str):
        if _valid_command(AttrDataFormat.SPECTRUM, dtype):
            self.add_command(
                command(
                    f=getattr(self, f"{name}_spectrum_cmd"),
                    dtype_in=dtype,
                    dtype_out=dtype,
                    dformat_in=AttrDataFormat.SPECTRUM,
                    dformat_out=AttrDataFormat.SPECTRUM,
                ),
            )

    def initialize_dynamic_attributes(self):
        for attr_data in _all_attribute_definitions:
            self.add_scalar_attr(
                attr_data.name, attr_data.tango_type, attr_data.initial_scalar
            )
            self.add_array_attrs(
                attr_data.name, attr_data.tango_type, attr_data.initial_spectrum
            )
            self.add_scalar_command(attr_data.name, attr_data.tango_type)
            self.add_spectrum_command(attr_data.name, attr_data.tango_type)

    @command
    def reset_values(self):
        for attr_name in self.attr_values:
            self.attr_values[attr_name] = self.initial_values[attr_name]

    def read(self, attr):
        value = self.attr_values[attr.get_name()]
        attr.set_value(value)

    def write(self, attr):
        new_value = attr.get_write_value()
        self.attr_values[attr.get_name()] = new_value
        self.push_change_event(attr.get_name(), new_value)

    echo_command_code = textwrap.dedent(
        """\
            def {}(self, arg):
                return arg
            """
    )

    for attr_data in _all_attribute_definitions:
        if _valid_command(AttrDataFormat.SCALAR, attr_data.tango_type):
            exec(echo_command_code.format(f"{attr_data.name}_cmd"))
        if _valid_command(AttrDataFormat.SPECTRUM, attr_data.tango_type):
            exec(echo_command_code.format(f"{attr_data.name}_spectrum_cmd"))



from collections.abc import AsyncGenerator, AsyncIterator

import numpy as np
from bluesky.protocols import Hints, StreamAsset
from event_model import DataKey

from ophyd_async.core import (
    DetectorWriter,
    HDFDatasetDescription,
    HDFDocumentComposer,
    PathProvider,
)

from ._pattern_generator import DATA_PATH, SUM_PATH, PatternGenerator

WIDTH = 320
HEIGHT = 240


class BlobDetectorWriter(DetectorWriter):
    def __init__(
        self,
        pattern_generator: PatternGenerator,
        path_provider: PathProvider,
    ) -> None:
        self.pattern_generator = pattern_generator
        self.path_provider = path_provider
        self.composer: HDFDocumentComposer | None = None
        self.datasets: list[HDFDatasetDescription] = []

    async def open(self, name: str, exposures_per_event: int = 1) -> dict[str, DataKey]:
        path_info = self.path_provider(name)
        write_path = path_info.directory_path / f"{path_info.filename}.h5"
        read_path_uri = f"{path_info.directory_uri}{path_info.filename}.h5"
        self.pattern_generator.open_file(write_path, WIDTH, HEIGHT)
        self.exposures_per_event = exposures_per_event
        # We know it will write data and sum, so emit those
        self.datasets = [
            HDFDatasetDescription(
                data_key=name,
                dataset=DATA_PATH,
                shape=(exposures_per_event, HEIGHT, WIDTH),
                dtype_numpy=np.dtype(np.uint8).str,
                chunk_shape=(1024,),
            ),
            HDFDatasetDescription(
                data_key=f"{name}-sum",
                dataset=SUM_PATH,
                shape=(exposures_per_event,) if exposures_per_event > 1 else (),
                dtype_numpy=np.dtype(np.int64).str,
                chunk_shape=(1024,),
            ),
        ]
        self.composer = HDFDocumentComposer(read_path_uri, self.datasets)
        describe = {
            ds.data_key: DataKey(
                source="sim://pattern-generator-hdf-file",
                shape=list(ds.shape),
                dtype="array"
                if exposures_per_event > 1 or len(ds.shape) > 1
                else "number",
                dtype_numpy=ds.dtype_numpy,
                external="STREAM:",
            )
            for ds in self.datasets
        }
        return describe

    def get_hints(self, name: str) -> Hints:
        """The hints to be used for the detector."""
        return {"fields": [name]}

    async def get_indices_written(self) -> int:
        return self.pattern_generator.get_last_index() // self.exposures_per_event

    async def observe_indices_written(
        self, timeout: float
    ) -> AsyncGenerator[int, None]:
        while True:
            yield self.pattern_generator.get_last_index() // self.exposures_per_event
            await self.pattern_generator.wait_for_next_index(timeout)

    async def collect_stream_docs(
        self, name: str, indices_written: int
    ) -> AsyncIterator[StreamAsset]:
        # When we have written something to the file
        if self.composer is None:
            msg = f"open() not called on {self}"
            raise RuntimeError(msg)
        for doc in self.composer.make_stream_docs(indices_written):
            yield doc

    async def close(self) -> None:
        self.pattern_generator.close_file()



import random
import string
import subprocess
import sys
import time
from pathlib import Path


def generate_random_pv_prefix() -> str:
    """For generating random PV names in test devices."""
    return "".join(random.choice(string.ascii_lowercase) for _ in range(12)) + ":"


class TestingIOC:
    """For initialising an IOC in tests."""

    def __init__(self):
        self._db_macros: list[tuple[Path, dict[str, str]]] = []
        self.output = ""

    def add_database(self, db: Path | str, /, **macros: str):
        self._db_macros.append((Path(db), macros))

    def start(self):
        ioc_args = [
            sys.executable,
            "-m",
            "epicscorelibs.ioc",
        ]
        for db, macros in self._db_macros:
            macro_str = ",".join(f"{k}={v}" for k, v in macros.items())
            ioc_args += ["-m", macro_str, "-d", str(db)]
        self._process = subprocess.Popen(
            ioc_args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        assert self._process.stdout  # noqa: S101 # this is to make Pylance happy
        start_time = time.monotonic()
        while "iocRun: All initialization complete" not in self.output:
            if time.monotonic() - start_time > 15:
                self.stop()
                raise TimeoutError(f"IOC did not start in time:\n{self.output}")
            self.output += self._process.stdout.readline()

    def stop(self):
        try:
            self.output += self._process.communicate("exit()")[0]
        except ValueError:
            # Someone else already called communicate
            pass



import logging
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest
from bluesky.run_engine import RunEngine

from ophyd_async.core import DetectorTrigger, TriggerInfo, init_devices
from ophyd_async.fastcs.jungfrau import AcquisitionType, DetectorStatus, Jungfrau
from ophyd_async.fastcs.jungfrau._signals import (
    JungfrauTriggerMode,  # noqa: PLC2701
    PedestalMode,  # noqa: PLC2701
)
from ophyd_async.testing import (
    callback_on_mock_put,
    set_mock_value,
)


@pytest.fixture
def jungfrau(RE: RunEngine):
    with init_devices(mock=True):
        detector = Jungfrau("prefix", MagicMock(), "", "", 4, "jungfrau")

    def set_meta_filename_and_id(value, *args, **kwargs):
        set_mock_value(detector.odin.meta_file_name, value)
        set_mock_value(detector.odin.id, value)

    callback_on_mock_put(detector.odin.file_name, set_meta_filename_and_id)

    detector._writer._path_provider.return_value.filename = "filename.h5"  # type: ignore

    set_mock_value(detector.odin.meta_active, "Active")
    set_mock_value(detector.odin.capture_rbv, "Capturing")
    set_mock_value(detector.odin.meta_writing, "Writing")
    return detector


@pytest.mark.parametrize(
    "trigger_info, match",
    [
        (
            TriggerInfo(trigger=DetectorTrigger.CONSTANT_GATE, deadtime=1),
            "The trigger method can only be called with internal or edge triggering",
        ),
        (
            TriggerInfo(),
            "Must set TriggerInfo.livetime",
        ),
        (
            TriggerInfo(trigger=DetectorTrigger.INTERNAL, number_of_events=10),
            "Number of events must be set to 1 in internal trigger mode during "
            "standard acquisitions.",
        ),
        (
            TriggerInfo(
                trigger=DetectorTrigger.EDGE_TRIGGER,
                exposures_per_event=10,
                deadtime=1,
                livetime=2,
            ),
            "Exposures per event must be set to 1 in edge trigger mode "
            "during standard acquisitions.",
        ),
    ],
)
async def test_prepare_val_error_on_bad_trigger_info(
    trigger_info: TriggerInfo, match: str, jungfrau: Jungfrau
):
    with pytest.raises(
        ValueError,
        match=match,
    ):
        await jungfrau.prepare(trigger_info)


async def test_prepare_warn_on_small_exposure(jungfrau: Jungfrau, caplog):
    bad_trigger_info = TriggerInfo(livetime=1e-6)
    with caplog.at_level(logging.WARNING):
        await jungfrau.prepare(bad_trigger_info)
    assert "Exposure time shorter than 2μs is not recommended" in caplog.messages


async def test_prepare_error_on_bad_no_of_event(
    jungfrau: Jungfrau,
):
    bad_trigger_info = TriggerInfo(number_of_events=[2], livetime=1e-3)
    with pytest.raises(TypeError, match="Number of events must be an integer"):
        await jungfrau.prepare(bad_trigger_info)


async def test_good_prepare(jungfrau: Jungfrau):
    good_trigger_info = TriggerInfo(livetime=1e-3)
    await jungfrau.prepare(good_trigger_info)


async def test_arm(jungfrau: Jungfrau):
    jungfrau.drv.acquisition_start.trigger = AsyncMock()
    await jungfrau._controller.arm()
    jungfrau.drv.acquisition_start.trigger.assert_called_once()


@patch("ophyd_async.fastcs.jungfrau._controller.wait_for_value")
async def test_wait_for_idle(mock_wait_for_value: AsyncMock, jungfrau: Jungfrau):
    await jungfrau._controller.wait_for_idle()
    mock_wait_for_value.assert_called_once_with(
        jungfrau.drv.detector_status, DetectorStatus.IDLE, timeout=10.0
    )


async def test_disarm(jungfrau: Jungfrau):
    jungfrau.drv.acquisition_stop.trigger = AsyncMock()
    jungfrau.drv.pedestal_mode_state.set = AsyncMock()
    jungfrau.drv.acquisition_type.set = AsyncMock()
    await jungfrau._controller.disarm()
    jungfrau.drv.acquisition_stop.trigger.assert_called_once()
    jungfrau.drv.pedestal_mode_state.set.assert_awaited_once_with(PedestalMode.OFF)
    jungfrau.drv.acquisition_type.set.assert_awaited_once_with(AcquisitionType.STANDARD)


async def test_signals_set_in_pedestal_mode(jungfrau: Jungfrau):
    await jungfrau.drv.acquisition_type.set(AcquisitionType.PEDESTAL)
    frames_and_events = 10
    good_trigger_info = TriggerInfo(
        livetime=1e-3,
        number_of_events=frames_and_events * 2,
        exposures_per_event=frames_and_events,
        trigger=DetectorTrigger.INTERNAL,
    )
    await jungfrau.prepare(good_trigger_info)
    assert await jungfrau.drv.pedestal_mode_frames.get_value() == frames_and_events
    assert await jungfrau.drv.pedestal_mode_loops.get_value() == frames_and_events


async def test_signals_set_in_standard_internal_mode(jungfrau: Jungfrau):
    jungfrau.drv.pedestal_mode_frames = AsyncMock()
    jungfrau.drv.pedestal_mode_loops = AsyncMock()
    exp_per_event = 10
    good_trigger_info = TriggerInfo(
        livetime=1e-3,
        exposures_per_event=exp_per_event,
        trigger=DetectorTrigger.INTERNAL,
    )
    await jungfrau.prepare(good_trigger_info)
    assert await jungfrau.drv.frames_per_acq.get_value() == exp_per_event
    jungfrau.drv.pedestal_mode_frames.assert_not_called()
    jungfrau.drv.pedestal_mode_frames.assert_not_called()


async def test_signals_set_in_standard_external_mode(jungfrau: Jungfrau):
    jungfrau.drv.pedestal_mode_frames = AsyncMock()
    jungfrau.drv.pedestal_mode_loops = AsyncMock()
    total_events = 10
    good_trigger_info = TriggerInfo(
        livetime=1e-3,
        number_of_events=total_events,
        trigger=DetectorTrigger.EDGE_TRIGGER,
        deadtime=1,
    )
    await jungfrau.prepare(good_trigger_info)
    assert await jungfrau.drv.frames_per_acq.get_value() == total_events
    jungfrau.drv.pedestal_mode_frames.assert_not_called()
    jungfrau.drv.pedestal_mode_frames.assert_not_called()


async def test_error_in_pedestal_and_external_modes(jungfrau: Jungfrau):
    await jungfrau.drv.acquisition_type.set(AcquisitionType.PEDESTAL)
    with pytest.raises(
        ValueError,
        match="Jungfrau must be triggered internally while in pedestal mode.",
    ):
        await jungfrau.prepare(
            TriggerInfo(livetime=1e-3, trigger=DetectorTrigger.EDGE_TRIGGER, deadtime=1)
        )


async def test_prepare_val_error_if_pedestal_mode_and_odd_number_of_events(
    jungfrau: Jungfrau,
):
    # No. events = pedestal loops * 2, so it should always be even
    await jungfrau.drv.acquisition_type.set(AcquisitionType.PEDESTAL)

    trigger_info = TriggerInfo(livetime=1e-3, deadtime=1)

    with patch("ophyd_async.core._signal.SignalW.set", return_value=None):
        with pytest.raises(
            ValueError,
            match=f"Invalid trigger info for pedestal mode. "
            f"{trigger_info.number_of_events=} must be divisible by two. "
            f"Was create_jungfrau_pedestal_triggering_info used?",
        ):
            await jungfrau.prepare(trigger_info)


async def test_prepare_pedestal_mode_sets_trigger_mode_before_pedestal_mode(
    jungfrau: Jungfrau,
):
    await jungfrau.drv.acquisition_type.set(AcquisitionType.PEDESTAL)
    trigger_info = TriggerInfo(
        livetime=1e-3,
        number_of_events=4,
        deadtime=1,
    )
    parent_mock = MagicMock()
    jungfrau.drv.pedestal_mode_state.set = AsyncMock()
    jungfrau.drv.trigger_mode.set = AsyncMock()
    parent_mock.attach_mock(jungfrau.drv.pedestal_mode_state.set, "pedestal_mode_state")
    parent_mock.attach_mock(jungfrau.drv.trigger_mode.set, "trigger_mode")
    await jungfrau.prepare(trigger_info)
    assert parent_mock.mock_calls == [
        call.trigger_mode(JungfrauTriggerMode.INTERNAL),
        call.pedestal_mode_state(PedestalMode.ON),
    ]



"""Support for the ADPilatus areaDetector driver.

https://github.com/areaDetector/ADPilatus
"""

from ._pilatus import PilatusDetector
from ._pilatus_controller import PilatusController, PilatusReadoutTime
from ._pilatus_io import PilatusDriverIO, PilatusTriggerMode

__all__ = [
    "PilatusDetector",
    "PilatusReadoutTime",
    "PilatusController",
    "PilatusDriverIO",
    "PilatusTriggerMode",
]



import asyncio
from enum import Enum

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    DetectorTrigger,
    TriggerInfo,
    wait_for_value,
)
from ophyd_async.epics import adcore

from ._pilatus_io import PilatusDriverIO, PilatusTriggerMode


class PilatusReadoutTime(float, Enum):
    """Pilatus readout time per model in ms."""

    # Cite: https://media.dectris.com/User_Manual-PILATUS2-V1_4.pdf
    PILATUS2 = 2.28e-3

    # Cite: https://media.dectris.com/user-manual-pilatus3-2020.pdf
    PILATUS3 = 0.95e-3


class PilatusController(adcore.ADBaseController[PilatusDriverIO]):
    """`DetectorController` for a `PilatusDriverIO`."""

    _supported_trigger_types = {
        DetectorTrigger.INTERNAL: PilatusTriggerMode.INTERNAL,
        DetectorTrigger.CONSTANT_GATE: PilatusTriggerMode.EXT_ENABLE,
        DetectorTrigger.VARIABLE_GATE: PilatusTriggerMode.EXT_ENABLE,
        DetectorTrigger.EDGE_TRIGGER: PilatusTriggerMode.EXT_TRIGGER,
    }

    def __init__(
        self,
        driver: PilatusDriverIO,
        good_states: frozenset[adcore.ADState] = adcore.DEFAULT_GOOD_STATES,
        readout_time: float = PilatusReadoutTime.PILATUS3,
    ) -> None:
        super().__init__(driver, good_states=good_states)
        self._readout_time = readout_time

    def get_deadtime(self, exposure: float | None) -> float:
        return self._readout_time

    async def prepare(self, trigger_info: TriggerInfo):
        if trigger_info.livetime is not None:
            await self.set_exposure_time_and_acquire_period_if_supplied(
                trigger_info.livetime
            )
        await asyncio.gather(
            self.driver.trigger_mode.set(
                self._supported_trigger_types[trigger_info.trigger]
            ),
            self.driver.num_images.set(
                999_999
                if trigger_info.total_number_of_exposures == 0
                else trigger_info.total_number_of_exposures
            ),
            self.driver.image_mode.set(adcore.ADImageMode.MULTIPLE),
        )

    async def arm(self):
        # Standard arm the detector and wait for the acquire PV to be True
        self._arm_status = await self.start_acquiring_driver_and_ensure_status()

        # The pilatus has an additional PV that goes True when the camserver
        # is actually ready. Should wait for that too or we risk dropping
        # a frame
        await wait_for_value(
            self.driver.armed,
            True,
            timeout=DEFAULT_TIMEOUT,
        )



import pytest
from scanspec.core import Path
from scanspec.specs import Fly, Line

from ophyd_async.epics.motor import Motor
from ophyd_async.epics.pmac import (
    PmacIO,
)
from ophyd_async.epics.pmac._utils import (
    _PmacMotorInfo,  # noqa: PLC2701
    calculate_ramp_position_and_duration,  # noqa: PLC2701
)
from ophyd_async.testing import set_mock_value


async def test_calculate_ramp_position_and_duration(
    sim_motors: tuple[PmacIO, Motor, Motor],
):
    sim_pmac, sim_x_motor, sim_y_motor = sim_motors
    spec = Fly(1.0 @ (Line(sim_y_motor, 10, 12, 3) * ~Line(sim_x_motor, 1, 5, 5)))
    slice = Path(spec.calculate()).consume()

    motor_info = await _PmacMotorInfo.from_motors(sim_pmac, [sim_x_motor, sim_y_motor])

    ramp_up_pos, ramp_up_time = calculate_ramp_position_and_duration(
        slice, motor_info, True
    )
    ramp_down_pos, ramp_down_time = calculate_ramp_position_and_duration(
        slice, motor_info, False
    )

    assert ramp_up_pos[sim_x_motor] == 0.45
    assert ramp_up_pos[sim_y_motor] == 10
    assert ramp_up_time == 0.1
    assert ramp_down_pos[sim_x_motor] == 5.55
    assert ramp_down_pos[sim_y_motor] == 12
    assert ramp_down_time == 0.1


async def test_motor_info_from_motors(sim_motors: tuple[PmacIO, Motor, Motor]):
    sim_pmac, sim_x_motor, sim_y_motor = sim_motors
    motor_info = await _PmacMotorInfo.from_motors(sim_pmac, [sim_x_motor, sim_y_motor])
    expected_motor_info = _PmacMotorInfo(
        "CS1",
        1,
        {sim_x_motor: 6, sim_y_motor: 7},
        {sim_x_motor: 10, sim_y_motor: 20},
        {sim_x_motor: 5, sim_y_motor: 10},
    )
    assert motor_info == expected_motor_info


async def test_multiple_cs_port_raises_runtime_error(
    sim_motors: tuple[PmacIO, Motor, Motor],
):
    sim_pmac, sim_x_motor, sim_y_motor = sim_motors
    set_mock_value(
        sim_pmac.assignment[sim_pmac.motor_assignment_index[sim_x_motor]].cs_port, "CS2"
    )

    with pytest.raises(RuntimeError, match="multiple CS ports"):
        await _PmacMotorInfo.from_motors(sim_pmac, [sim_x_motor, sim_y_motor])


async def test_multiple_cs_number_raises_runtime_error(
    sim_motors: tuple[PmacIO, Motor, Motor],
):
    sim_pmac, sim_x_motor, sim_y_motor = sim_motors
    set_mock_value(
        sim_pmac.assignment[sim_pmac.motor_assignment_index[sim_x_motor]].cs_number, 2
    )

    with pytest.raises(RuntimeError, match="multiple CS numbers"):
        await _PmacMotorInfo.from_motors(sim_pmac, [sim_x_motor, sim_y_motor])


async def test_duplicate_cs_axis_letter_raises_runtime_error(
    sim_motors: tuple[PmacIO, Motor, Motor],
):
    sim_pmac, sim_x_motor, sim_y_motor = sim_motors
    set_mock_value(
        sim_pmac.assignment[
            sim_pmac.motor_assignment_index[sim_x_motor]
        ].cs_axis_letter,
        "Y",
    )

    with pytest.raises(RuntimeError, match="same CS Axis"):
        await _PmacMotorInfo.from_motors(sim_pmac, [sim_x_motor, sim_y_motor])


async def test_unexpected_cs_axis_letter_raises_value_error(
    sim_motors: tuple[PmacIO, Motor, Motor],
):
    sim_pmac, sim_x_motor, sim_y_motor = sim_motors
    set_mock_value(
        sim_pmac.assignment[
            sim_pmac.motor_assignment_index[sim_x_motor]
        ].cs_axis_letter,
        "I",
    )

    with pytest.raises(ValueError, match="Failed to get motor CS index"):
        await _PmacMotorInfo.from_motors(sim_pmac, [sim_x_motor, sim_y_motor])


async def test_blank_cs_axis_letter_raises_value_error(
    sim_motors: tuple[PmacIO, Motor, Motor],
):
    sim_pmac, sim_x_motor, sim_y_motor = sim_motors
    set_mock_value(
        sim_pmac.assignment[
            sim_pmac.motor_assignment_index[sim_x_motor]
        ].cs_axis_letter,
        "",
    )

    with pytest.raises(ValueError, match="Failed to get motor CS index"):
        await _PmacMotorInfo.from_motors(sim_pmac, [sim_x_motor, sim_y_motor])



from collections.abc import Sequence
from typing import Any

import numpy as np
import pytest
from pydantic import ValidationError

from ophyd_async.core import Array1D, Table


class MyTable(Table):
    bool: Array1D[np.bool_]
    uint: Array1D[np.uint32]
    str: Sequence[str]


@pytest.mark.parametrize(
    ["kwargs", "error_msg"],
    [
        (
            {"bool": [3, 4], "uint": [3, 4], "str": ["", ""]},
            "bool: Cannot cast [3, 4] to bool without losing precision",
        ),
        (
            {"bool": np.array([1], dtype=np.uint8), "uint": [-3], "str": [""]},
            "uint: Cannot cast [-3] to uint32 without losing precision",
        ),
        (
            {"bool": [0], "uint": np.array([1.8], dtype=np.float64), "str": [""]},
            "uint: Cannot cast [1.8] to uint32 without losing precision",
        ),
        (
            {"bool": [0, 1], "uint": [3, 4], "str": [44, ""]},
            "Input should be a valid string [type=string_type, input_value=44,",
        ),
        (
            {"bool": [0, 1], "uint": [[3], [4]], "str": ["", ""]},
            "Array 2-dimensional; the target dimensions is 1",
        ),
    ],
)
def test_table_wrong_types(kwargs, error_msg):
    with pytest.raises(ValidationError) as exc:
        MyTable(**kwargs)
    assert error_msg in str(exc.value)


@pytest.mark.parametrize(
    "kwargs",
    [
        {"bool": np.array([1], dtype=np.uint8), "uint": [3], "str": ["a"]},
        {
            "bool": np.array([False], dtype=np.bool_),
            "uint": np.array([1], dtype=np.float64),
            "str": ["b"],
        },
    ],
)
def test_table_coerces(kwargs):
    t = MyTable(**kwargs)
    for k, v in t:
        assert v == pytest.approx(kwargs[k])
    assert t == pytest.approx(t)


def test_validate_array_dtypes():
    class TestTable(Table):
        int_array: np.ndarray[Any, np.dtype[np.int32]]

    with pytest.raises(ValueError, match=r"Cannot cast .* without losing precision"):
        TestTable(int_array=[1.5, 2.5])  # type: ignore

    table = TestTable(int_array=[1, 2])  # type: ignore
    assert table.int_array.dtype == np.dtype(np.int32)

    table = TestTable(int_array=[])  # type: ignore
    assert table.int_array.dtype == np.dtype(np.int32)



from collections.abc import Sequence

from ophyd_async.core import PathProvider, SignalR, StandardDetector

from ._core_io import ADBaseIO, NDPluginBaseIO, NDPluginCBIO
from ._core_logic import ADBaseContAcqController, ADBaseControllerT
from ._core_writer import ADWriter
from ._hdf_writer import ADHDFWriter


class AreaDetector(StandardDetector[ADBaseControllerT, ADWriter]):
    def __init__(
        self,
        controller: ADBaseControllerT,
        writer: ADWriter,
        plugins: dict[str, NDPluginBaseIO] | None = None,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
    ):
        self.driver = controller.driver
        self.fileio = writer.fileio

        if plugins is not None:
            for plugin_name, plugin in plugins.items():
                setattr(self, plugin_name, plugin)

        super().__init__(
            controller,
            writer,
            (self.driver.acquire_period, self.driver.acquire_time, *config_sigs),
            name=name,
        )

    def get_plugin(
        self, name: str, plugin_type: type[NDPluginBaseIO] = NDPluginBaseIO
    ) -> NDPluginBaseIO:
        plugin = getattr(self, name, None)
        if not isinstance(plugin, plugin_type):
            raise TypeError(
                f"Expected {self.name}.{name} to be a {plugin_type}, got {plugin}"
            )
        return plugin


class ContAcqAreaDetector(AreaDetector[ADBaseContAcqController]):
    """Ophyd-async implementation of a continuously acquiring AreaDetector."""

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider,
        drv_cls: type[ADBaseIO] = ADBaseIO,
        drv_suffix: str = "cam1:",
        cb_suffix: str = "CB1:",
        writer_cls: type[ADWriter] = ADHDFWriter,
        fileio_suffix: str | None = None,
        name: str = "",
        plugins: dict[str, NDPluginBaseIO] | None = None,
        config_sigs: Sequence[SignalR] = (),
    ):
        self.cb_plugin = NDPluginCBIO(prefix + cb_suffix)
        driver = drv_cls(prefix + drv_suffix)
        controller = ADBaseContAcqController(driver, self.cb_plugin)

        writer = writer_cls.with_io(
            prefix,
            path_provider,
            # Since the CB plugin controls acq, use it when checking shape
            dataset_source=self.cb_plugin,
            fileio_suffix=fileio_suffix,
            plugins=plugins,
        )

        super().__init__(
            controller=controller,
            writer=writer,
            plugins=plugins,
            name=name,
            config_sigs=config_sigs,
        )



from ophyd_async.core import DetectorTrigger, TriggerInfo
from ophyd_async.fastcs.jungfrau import (
    JUNGFRAU_DEADTIME_S,
    create_jungfrau_external_triggering_info,
    create_jungfrau_internal_triggering_info,
    create_jungfrau_pedestal_triggering_info,
)


def test_create_jungfrau_internal_triggering_info():
    assert create_jungfrau_internal_triggering_info(5, 1) == TriggerInfo(
        number_of_events=1,
        deadtime=0,
        livetime=1,
        exposures_per_event=5,
    )


def test_create_jungfrau_external_triggering_info():
    assert create_jungfrau_external_triggering_info(
        total_triggers=5,
        exposure_time_s=0.01,
    ) == TriggerInfo(
        number_of_events=5,
        trigger=DetectorTrigger.EDGE_TRIGGER,
        livetime=0.01,
        deadtime=JUNGFRAU_DEADTIME_S,
    )


def test_create_external_triggering_info_regular_deadtime_if_period_not_specified():
    assert create_jungfrau_external_triggering_info(
        total_triggers=5,
        exposure_time_s=0.01,
    ) == TriggerInfo(
        number_of_events=5,
        trigger=DetectorTrigger.EDGE_TRIGGER,
        deadtime=JUNGFRAU_DEADTIME_S,
        livetime=0.01,
    )


async def test_create_jungfrau_pedestal_triggering_info():
    assert create_jungfrau_pedestal_triggering_info(
        exposure_time_s=0.01, pedestal_frames=5, pedestal_loops=10
    ) == TriggerInfo(
        trigger=DetectorTrigger.INTERNAL,
        number_of_events=20,
        exposures_per_event=5,
        livetime=0.01,
    )



"""Plan stubs for connecting, setting up and flying devices."""

from ._ensure_connected import ensure_connected
from ._nd_attributes import setup_ndattributes, setup_ndstats_sum
from ._panda import apply_panda_settings
from ._settings import (
    apply_settings,
    apply_settings_if_different,
    get_current_settings,
    retrieve_settings,
    store_settings,
)

__all__ = [
    "ensure_connected",
    "setup_ndattributes",
    "setup_ndstats_sum",
    "apply_panda_settings",
    "apply_settings",
    "apply_settings_if_different",
    "get_current_settings",
    "retrieve_settings",
    "store_settings",
]



"""Integration tests for a StandardDetector using a ADHDFWriter and SimController."""

import os
import time
from collections import defaultdict
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import cast
from unittest.mock import patch

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import pytest
from bluesky.run_engine import RunEngine
from bluesky.utils import FailedStatus

import ophyd_async.plan_stubs as ops
from ophyd_async.core import (
    AsyncStatus,
    DetectorTrigger,
    StaticFilenameProvider,
    StaticPathProvider,
    TriggerInfo,
)
from ophyd_async.epics import adcore, adsimdetector
from ophyd_async.testing import assert_emitted, set_mock_value


@pytest.fixture
def test_adsimdetector(ad_standard_det_factory: Callable) -> adsimdetector.SimDetector:
    return ad_standard_det_factory(adsimdetector.SimDetector)


@pytest.fixture
def test_adsimdetector_tiff(
    ad_standard_det_factory: Callable,
) -> adsimdetector.SimDetector:
    return ad_standard_det_factory(adsimdetector.SimDetector, adcore.ADTIFFWriter)


@pytest.fixture
def two_test_adsimdetectors(
    ad_standard_det_factory: Callable,
) -> Sequence[adsimdetector.SimDetector]:
    deta = ad_standard_det_factory(adsimdetector.SimDetector)
    detb = ad_standard_det_factory(adsimdetector.SimDetector, number=2)

    return deta, detb


def count_sim(
    dets: Sequence[adsimdetector.SimDetector],
    times: int = 1,
    trigger_info: TriggerInfo | None = None,
):
    """Test plan to do the equivalent of bp.count for a sim detector."""

    yield from bps.stage_all(*dets)
    yield from bps.open_run()
    if trigger_info:
        for det in dets:
            yield from bps.prepare(det, trigger_info, wait=True)
    for _ in range(times):
        read_values = {}
        for det in dets:
            read_values[det] = yield from bps.rd(det._writer.fileio.num_captured)

        for det in dets:
            yield from bps.trigger(det, wait=False, group="wait_for_trigger")

        yield from bps.sleep(1.0)

        # Assume that the number of images configured is the number of images captured
        for det in dets:
            num_images = yield from bps.rd(det.driver.num_images)
            set_mock_value(
                det._writer.fileio.num_captured,
                read_values[det] + num_images,
            )

        yield from bps.wait(group="wait_for_trigger")
        yield from bps.create()

        for det in dets:
            yield from bps.read(det)

        yield from bps.save()

    yield from bps.close_run()
    yield from bps.unstage_all(*dets)


@pytest.mark.timeout(3.5)
async def test_detector_count_failure(
    test_adsimdetector: adsimdetector.SimDetector,
    RE: RunEngine,
):
    """Preparing a step scan to use more than one event fails.

    Step scans always produce one event.
    """
    trigger_info = TriggerInfo(
        number_of_events=10,
        trigger=DetectorTrigger.INTERNAL,
        exposures_per_event=5,
    )
    try:
        with pytest.raises(FailedStatus) as exc:
            RE(count_sim([test_adsimdetector], times=1, trigger_info=trigger_info))
        assert isinstance(exc.value.__cause__, ValueError)
    finally:
        RE(bps.unstage(test_adsimdetector, wait=True))


@pytest.mark.timeout(7.5)
@pytest.mark.parametrize("exposures_per_event", [1, 5])
async def test_detector_count(
    test_adsimdetector: adsimdetector.SimDetector,
    RE: RunEngine,
    exposures_per_event: int,
):
    docs = defaultdict(list)
    RE.subscribe(lambda name, doc: docs[name].append(doc))
    trigger_info = TriggerInfo(
        number_of_events=1,
        trigger=DetectorTrigger.INTERNAL,
        exposures_per_event=exposures_per_event,
    )
    RE(count_sim([test_adsimdetector], times=5, trigger_info=trigger_info))

    assert_emitted(
        docs,
        start=1,
        descriptor=1,
        stream_resource=1,
        stream_datum=5,
        event=5,
        stop=1,
    )


@pytest.mark.parametrize("exposures_per_event", [1, 5, 15])
async def test_detector_fly(
    test_adsimdetector: adsimdetector.SimDetector,
    RE: RunEngine,
    exposures_per_event: int,
):
    trigger_info = TriggerInfo(
        number_of_events=15,
        trigger=DetectorTrigger.INTERNAL,
        exposures_per_event=exposures_per_event,
    )
    docs = defaultdict(list)
    RE.subscribe(lambda name, doc: docs[name].append(doc))

    def assert_n_stream_datums(
        n: int, start: int | None = None, stop: int | None = None
    ):
        if n == 0:
            assert "stream_datum" not in docs
        else:
            assert len(docs["stream_datum"]) == n
            # check both detectors have the correct start/stop
            seq_nums = docs["stream_datum"][n - 1]["seq_nums"]
            assert seq_nums["start"] == start
            assert seq_nums["stop"] == stop

    @bpp.stage_decorator([test_adsimdetector])
    @bpp.run_decorator()
    def fly_plan():
        yield from bps.prepare(test_adsimdetector, trigger_info, wait=True)
        yield from bps.declare_stream(test_adsimdetector, name="primary")
        yield from bps.kickoff(test_adsimdetector, wait=True)
        yield from bps.complete(test_adsimdetector, wait=False, group="complete_fly")

        # Don't process a full event yet and ensure no stream datum is emitted
        set_mock_value(
            test_adsimdetector.fileio.num_captured, trigger_info.exposures_per_event - 1
        )
        yield from bps.collect(test_adsimdetector)
        assert_n_stream_datums(0)

        # Process a full event and emit a stream datum
        set_mock_value(
            test_adsimdetector.fileio.num_captured, trigger_info.exposures_per_event
        )
        yield from bps.collect(test_adsimdetector)
        assert_n_stream_datums(1, 1, 2)

        if trigger_info.exposures_per_event > 1:
            # Process a full event + 1 exposure and make sure no stream datum is emitted
            set_mock_value(
                test_adsimdetector.fileio.num_captured, trigger_info.exposures_per_event
            )
            yield from bps.collect(test_adsimdetector)
            assert_n_stream_datums(1, 1, 2)

        # Process three full events and emit three stream data
        set_mock_value(
            test_adsimdetector.fileio.num_captured, 3 * trigger_info.exposures_per_event
        )
        yield from bps.collect(test_adsimdetector)
        assert_n_stream_datums(2, 2, 4)

        set_mock_value(
            test_adsimdetector.fileio.num_captured, 5 * trigger_info.exposures_per_event
        )
        yield from bps.collect(test_adsimdetector)
        assert_n_stream_datums(3, 4, 6)

        set_mock_value(
            test_adsimdetector.fileio.num_captured, 7 * trigger_info.exposures_per_event
        )
        yield from bps.collect(test_adsimdetector)
        assert_n_stream_datums(4, 6, 8)

        set_mock_value(
            test_adsimdetector.fileio.num_captured,
            15 * trigger_info.exposures_per_event,
        )
        yield from bps.collect(test_adsimdetector)
        assert_n_stream_datums(5, 8, 16)

        yield from bps.wait(group="complete_fly")

    RE(fly_plan())
    assert_emitted(
        docs, start=1, descriptor=1, stream_resource=1, stream_datum=5, stop=1
    )


async def test_two_detectors_fly_different_rate(
    two_test_adsimdetectors: Sequence[adsimdetector.SimDetector], RE: RunEngine
):
    docs = defaultdict(list)
    RE.subscribe(lambda name, doc: docs[name].append(doc))

    def assert_n_stream_datums(
        n: int, start: int | None = None, stop: int | None = None
    ):
        if n == 0:
            assert "stream_datum" not in docs
        else:
            assert len(docs["stream_datum"]) == n
            # check both detectors have the correct start/stop
            for detector_index in {n - 1, n - 2}:
                seq_nums = docs["stream_datum"][detector_index]["seq_nums"]
                assert seq_nums["start"] == start
                assert seq_nums["stop"] == stop

    @bpp.stage_decorator(two_test_adsimdetectors)
    @bpp.run_decorator()
    def fly_plan():
        for i, det in enumerate(two_test_adsimdetectors):
            yield from bps.prepare(
                det,
                TriggerInfo(
                    number_of_events=15,
                    trigger=DetectorTrigger.INTERNAL,
                    exposures_per_event=i + 1,
                ),
                wait=True,
                group="prepare",
            )
        yield from bps.declare_stream(*two_test_adsimdetectors, name="primary")

        for det in two_test_adsimdetectors:
            yield from bps.kickoff(det, wait=True)
        for det in two_test_adsimdetectors:
            yield from bps.complete(det, wait=False, group="complete_cleanup")

        # det[0] captures 5 frames, but we do not emit a StreamDatum as det[1] has not
        set_mock_value(two_test_adsimdetectors[0].fileio.num_captured, 5)

        yield from bps.collect(*two_test_adsimdetectors)
        assert_n_stream_datums(0)

        # det[0] captures 10 frames, but we do not emit a StreamDatum as det[1] has not
        set_mock_value(two_test_adsimdetectors[0].fileio.num_captured, 10)
        yield from bps.collect(*two_test_adsimdetectors)
        assert_n_stream_datums(0)

        # det[1] has caught up to first 7 frames, emit streamDatum for seq_num {1,4}
        set_mock_value(two_test_adsimdetectors[1].fileio.num_captured, 7)
        yield from bps.collect(*two_test_adsimdetectors)
        assert_n_stream_datums(2, 1, 4)

        # det[1] has caught up to first 14 frames, emit streamDatum for seq_num {4,8}
        set_mock_value(two_test_adsimdetectors[1].fileio.num_captured, 14)
        yield from bps.collect(*two_test_adsimdetectors)
        assert_n_stream_datums(4, 4, 8)

        # Complete both detectors
        for i, det in enumerate(two_test_adsimdetectors):
            set_mock_value(det.fileio.num_captured, 15 * (i + 1))

        # emits stream datum for seq_num {8, 15}
        yield from bps.collect(*two_test_adsimdetectors)
        assert_n_stream_datums(6, 8, 16)

        # Trigger has complete as all expected frames written
        yield from bps.wait("trigger_cleanup")

    RE(fly_plan())
    assert_emitted(
        docs, start=1, descriptor=1, stream_resource=2, stream_datum=6, stop=1
    )


@pytest.mark.timeout(3.5)
async def test_two_detectors_step(
    two_test_adsimdetectors: list[adsimdetector.SimDetector],
    RE: RunEngine,
):
    names = []
    docs = []
    RE.subscribe(lambda name, _: names.append(name))
    RE.subscribe(lambda _, doc: docs.append(doc))
    [
        set_mock_value(det.fileio.file_path_exists, True)
        for det in two_test_adsimdetectors
    ]

    controller_a = two_test_adsimdetectors[0]._controller
    writer_a = two_test_adsimdetectors[0]._writer
    writer_b = two_test_adsimdetectors[1]._writer
    info_a = writer_a._path_provider(device_name=two_test_adsimdetectors[0].name)
    info_b = writer_b._path_provider(device_name=two_test_adsimdetectors[1].name)
    file_name_a = None
    file_name_b = None

    def plan():
        nonlocal file_name_a, file_name_b
        yield from count_sim(two_test_adsimdetectors, times=1)

        drv = controller_a.driver
        assert False is (yield from bps.rd(drv.acquire))
        assert adcore.ADImageMode.MULTIPLE == (yield from bps.rd(drv.image_mode))

        hdfb = cast(adcore.NDFileHDFIO, writer_b.fileio)
        assert True is (yield from bps.rd(hdfb.lazy_open))
        assert True is (yield from bps.rd(hdfb.swmr_mode))
        assert 0 == (yield from bps.rd(hdfb.num_capture))
        assert adcore.ADFileWriteMode.STREAM == (
            yield from bps.rd(hdfb.file_write_mode)
        )

        assert (yield from bps.rd(writer_a.fileio.file_path)) == str(
            info_a.directory_path
        ) + os.sep
        file_name_a = yield from bps.rd(writer_a.fileio.file_name)
        assert file_name_a == info_a.filename

        assert (yield from bps.rd(writer_b.fileio.file_path)) == str(
            info_b.directory_path
        ) + os.sep
        file_name_b = yield from bps.rd(writer_b.fileio.file_name)
        assert file_name_b == info_b.filename

    RE(plan())
    assert names == [
        "start",
        "descriptor",
        "stream_resource",
        "stream_datum",
        "stream_resource",
        "stream_datum",
        "event",
        "stop",
    ]

    _, descriptor, sra, sda, srb, sdb, event, _ = docs
    assert descriptor["configuration"]["test_adsim1"]["data"][
        "test_adsim1-driver-acquire_time"
    ] == pytest.approx(0.8)
    assert descriptor["configuration"]["test_adsim2"]["data"][
        "test_adsim2-driver-acquire_time"
    ] == pytest.approx(1.8)
    assert descriptor["data_keys"]["test_adsim1"]["shape"] == [1, 10, 10]
    assert descriptor["data_keys"]["test_adsim2"]["shape"] == [1, 11, 11]
    assert sda["stream_resource"] == sra["uid"]
    assert sdb["stream_resource"] == srb["uid"]
    assert (
        srb["uri"]
        == "file://localhost/"
        + (info_b.directory_path / info_b.filename).as_posix().lstrip("/")
        + ".h5"
    )
    assert (
        sra["uri"]
        == "file://localhost/"
        + (info_a.directory_path / info_a.filename).as_posix().lstrip("/")
        + ".h5"
    )

    assert event["data"] == {}


@pytest.mark.timeout(5.5)
@pytest.mark.parametrize(
    "writer_cls", [adcore.ADHDFWriter, adcore.ADTIFFWriter, adcore.ADJPEGWriter]
)
async def test_detector_writes_to_file(
    RE: RunEngine,
    ad_standard_det_factory: Callable,
    writer_cls: type[adcore.ADWriter],
    tmp_path: Path,
):
    test_adsimdetector: adsimdetector.SimDetector = ad_standard_det_factory(
        adsimdetector.SimDetector, writer_cls
    )

    names = []
    docs = []
    RE.subscribe(lambda name, _: names.append(name))
    RE.subscribe(lambda _, doc: docs.append(doc))
    set_mock_value(
        test_adsimdetector.fileio.file_path_exists,
        True,
    )

    RE(count_sim([test_adsimdetector], times=3))

    assert (
        await test_adsimdetector.fileio.file_path.get_value() == str(tmp_path) + os.sep
    )

    descriptor_index = names.index("descriptor")

    assert docs[descriptor_index].get("data_keys").get(test_adsimdetector.name).get(
        "shape"
    ) == [
        1,
        10,
        10,
    ]
    assert names == [
        "start",
        "descriptor",
        "stream_resource",
        "stream_datum",
        "event",
        "stream_datum",
        "event",
        "stream_datum",
        "event",
        "stop",
    ]


async def test_read_and_describe_detector(
    test_adsimdetector: adsimdetector.SimDetector,
):
    describe = await test_adsimdetector.describe_configuration()
    read = await test_adsimdetector.read_configuration()
    assert describe == {
        "test_adsim1-driver-acquire_time": {
            "source": "mock+ca://SIM1:cam1:AcquireTime_RBV",
            "dtype": "number",
            "dtype_numpy": "<f8",
            "shape": [],
        },
        "test_adsim1-driver-acquire_period": {
            "source": "mock+ca://SIM1:cam1:AcquirePeriod_RBV",
            "dtype": "number",
            "dtype_numpy": "<f8",
            "shape": [],
        },
    }
    assert read == {
        "test_adsim1-driver-acquire_time": {
            "value": 0.8,
            "timestamp": pytest.approx(time.time(), rel=1e-2),
            "alarm_severity": 0,
        },
        "test_adsim1-driver-acquire_period": {
            "value": 1.0,
            "timestamp": pytest.approx(time.time(), rel=1e-2),
            "alarm_severity": 0,
        },
    }


async def test_read_returns_nothing(test_adsimdetector: adsimdetector.SimDetector):
    assert await test_adsimdetector.read() == {}


async def test_trigger_logic():
    """I want this test to check that when StandardDetector.trigger is called:

    1. the detector.controller is armed, and that starts the acquisition so that,
    2. The detector.writer.hdf.num_captured is 1

    Probably the best thing to do here is mock the detector.controller.driver and
    detector.writer.hdf. Then, mock out set_and_wait_for_value in the SimController
    so that, as well as setting detector.controller.driver.acquire to True, it sets
    detector.writer.hdf.num_captured to 1, using set_mock_value
    """
    ...


@pytest.mark.parametrize(
    "driver_name, error_output",
    [
        ("", "config signal must be named before it is passed to the detector"),
        (
            "some-name",
            (
                "config signal some-name-acquire_time must be connected before it is "
                "passed to the detector"
            ),
        ),
    ],
)
def test_detector_with_unnamed_or_disconnected_config_sigs(
    RE,
    static_filename_provider: StaticFilenameProvider,
    tmp_path: Path,
    driver_name,
    error_output,
):
    dp = StaticPathProvider(static_filename_provider, tmp_path)

    some_other_driver = adsimdetector.SimDriverIO("TEST", name=driver_name)

    det = adsimdetector.SimDetector(
        "FOO:",
        dp,
        name="foo",
    )

    det._config_sigs = [some_other_driver.acquire_time, det.driver.acquire]

    def my_plan():
        yield from ops.ensure_connected(det, mock=True)
        assert det.driver.acquire.name == "foo-driver-acquire"
        assert some_other_driver.acquire_time.name == (
            driver_name + "-acquire_time" if driver_name else ""
        )

        yield from count_sim([det], times=1)

    with pytest.raises(Exception) as exc:
        RE(my_plan())

    assert isinstance(exc.value.args[0], AsyncStatus)
    assert str(exc.value.args[0].exception()) == error_output

    # Need to unstage to properly kill tasks
    RE(bps.unstage(det, wait=True))


async def test_ad_sim_controller(test_adsimdetector: adsimdetector.SimDetector):
    ad = test_adsimdetector._controller
    with patch("ophyd_async.core._signal.wait_for_value", return_value=None):
        await ad.prepare(
            TriggerInfo(number_of_events=1, trigger=DetectorTrigger.INTERNAL)
        )
        await ad.arm()
        await ad.wait_for_idle()

    driver = ad.driver
    assert await driver.num_images.get_value() == 1
    assert await driver.image_mode.get_value() == adcore.ADImageMode.MULTIPLE
    assert await driver.acquire.get_value() is True

    await ad.disarm()

    assert await driver.acquire.get_value() is False



from __future__ import annotations

import asyncio
from collections.abc import Sequence
from dataclasses import dataclass

import numpy as np
from scanspec.core import Slice

from ophyd_async.core import gather_dict
from ophyd_async.epics.motor import Motor

from ._pmac_io import CS_LETTERS, PmacIO

# PMAC durations are in milliseconds
# We must convert from scanspec durations (seconds) to milliseconds
# PMAC motion program multiples durations by 0.001
# (see https://github.com/DiamondLightSource/pmac/blob/afe81f8bb9179c3a20eff351f30bc6cfd1539ad9/pmacApp/pmc/trajectory_scan_code_ppmac.pmc#L241)
# Therefore, we must divide scanspec durations by 10e-6
TICK_S = 0.000001
MIN_TURNAROUND = 0.002
MIN_INTERVAL = 0.002


@dataclass
class _PmacMotorInfo:
    cs_port: str
    cs_number: int
    motor_cs_index: dict[Motor, int]
    motor_acceleration_rate: dict[Motor, float]
    motor_max_velocity: dict[Motor, float]

    @classmethod
    async def from_motors(cls, pmac: PmacIO, motors: Sequence[Motor]) -> _PmacMotorInfo:
        """Creates a _PmacMotorInfo instance based on a  controller and list of motors.

        :param pmac: The PMAC device
        :param motors: Sequence of motors involved in trajectory
        :raises RuntimeError:
            if motors do not share common CS port or CS number, or if
            motors do not have unique CS index assignments
        :returns:
            _PmacMotorInfo instance with motor's common CS port and CS number, and
            dictionaries of motor's to their unique CS index and accelerate rate

        """
        assignments = {
            motor: pmac.assignment[pmac.motor_assignment_index[motor]]
            for motor in motors
        }

        cs_ports, cs_numbers, cs_axes, velocities, accls = await asyncio.gather(
            gather_dict(
                {motor: assignments[motor].cs_port.get_value() for motor in motors}
            ),
            gather_dict(
                {motor: assignments[motor].cs_number.get_value() for motor in motors}
            ),
            gather_dict(
                {
                    motor: assignments[motor].cs_axis_letter.get_value()
                    for motor in motors
                }
            ),
            gather_dict({motor: motor.max_velocity.get_value() for motor in motors}),
            gather_dict(
                {motor: motor.acceleration_time.get_value() for motor in motors}
            ),
        )

        # check if the values in cs_port and cs_number are the same
        cs_ports = set(cs_ports.values())

        if len(cs_ports) != 1:
            raise RuntimeError(
                "Failed to fetch common CS port."
                "Motors passed are assigned to multiple CS ports:"
                f"{list(cs_ports)}"
            )

        cs_port = cs_ports.pop()

        cs_numbers = set(cs_numbers.values())
        if len(cs_numbers) != 1:
            raise RuntimeError(
                "Failed to fetch common CS number."
                "Motors passed are assigned to multiple CS numbers:"
                f"{list(cs_numbers)}"
            )

        cs_number = cs_numbers.pop()

        motor_cs_index = {}
        for motor in cs_axes:
            try:
                if not cs_axes[motor]:
                    raise ValueError
                motor_cs_index[motor] = CS_LETTERS.index(cs_axes[motor])
            except ValueError as err:
                raise ValueError(
                    "Failed to get motor CS index. "
                    f"Motor {motor.name} assigned to '{cs_axes[motor]}' "
                    f"but must be assignmed to '{CS_LETTERS}"
                ) from err
            if len(set(motor_cs_index.values())) != len(motor_cs_index.items()):
                raise RuntimeError(
                    "Failed to fetch distinct CS Axes."
                    "Motors passed are assigned to the same CS Axis"
                    f"{list(motor_cs_index)}"
                )

        motor_acceleration_rate = {
            motor: float(velocities[motor]) / float(accls[motor])
            for motor in velocities
        }

        return _PmacMotorInfo(
            cs_port, cs_number, motor_cs_index, motor_acceleration_rate, velocities
        )


def calculate_ramp_position_and_duration(
    slice: Slice[Motor], motor_info: _PmacMotorInfo, is_up: bool
) -> tuple[dict[Motor, np.float64], float]:
    """Calculate the the required ramp position and duration of a trajectory.

    This function will:
      - Calculate the ramp time required to achieve each motor's
        initial entry velocity into the first frame of a slice
        or final exit velocity out of the last frame of a slice
      - Use the longest ramp time to calculate all motor's
        ramp up positions.

    :param slice: Information about a series of scan frames along a number of axes
    :param motor_info: Instance of _PmacMotorInfo
    :param is_up: Boolean representing ramping up into a frame or down out of a frame
    :returns tuple: A tuple containing:
        dict: Motor to ramp positions
        float: Ramp time required for all motors
    """
    if slice.duration is None:
        raise ValueError("Slice must have a duration")

    scan_axes = slice.axes()
    idx = 0 if is_up else -1

    velocities: dict[Motor, float] = {}
    ramp_times: list[float] = []
    for axis in scan_axes:
        velocity = (slice.upper[axis][idx] - slice.lower[axis][idx]) / slice.duration[
            idx
        ]
        velocities[axis] = velocity
        ramp_times.append(abs(velocity) / motor_info.motor_acceleration_rate[axis])
    ramp_times.append(
        MIN_TURNAROUND
    )  # Adding a 2ms ramp time as a min tournaround time
    max_ramp_time = max(ramp_times)

    motor_to_ramp_position = {}
    sign = -1 if is_up else 1
    for axis, v in velocities.items():
        ref_pos = slice.lower[axis][0] if is_up else slice.upper[axis][-1]
        displacement = 0.5 * v * max_ramp_time
        motor_to_ramp_position[axis] = ref_pos + sign * displacement

    return (motor_to_ramp_position, max_ramp_time)



from unittest.mock import AsyncMock, MagicMock

import pytest

from ophyd_async.core import DetectorTrigger, TriggerInfo, init_devices
from ophyd_async.fastcs.eiger import EigerDetector
from ophyd_async.testing import callback_on_mock_put, get_mock_put, set_mock_value


@pytest.fixture
def detector(RE):
    with init_devices(mock=True):
        detector = EigerDetector("BL03I", MagicMock())

    def set_meta_filename_and_id(value, *args, **kwargs):
        set_mock_value(detector.odin.meta_file_name, value)
        set_mock_value(detector.odin.id, value)

    callback_on_mock_put(detector.odin.file_name, set_meta_filename_and_id)

    detector._writer._path_provider.return_value.filename = "filename.h5"  # type: ignore

    set_mock_value(detector.odin.meta_active, "Active")
    set_mock_value(detector.odin.capture_rbv, "Capturing")
    set_mock_value(detector.odin.meta_writing, "Writing")
    return detector


async def test_when_prepared_eiger_bit_depth_is_passed_and_set_in_odin(detector):
    detector._controller.arm = AsyncMock()
    expected_datatype = 16
    set_mock_value(detector.drv.detector.bit_depth_image, expected_datatype)

    await detector.prepare(
        TriggerInfo(
            exposure_timeout=None,
            number_of_events=1,
            trigger=DetectorTrigger.INTERNAL,
        )
    )

    # Assert that odin datatype is set to the eiger bit depth during detector prepare
    get_mock_put(detector.odin.data_type).assert_called_once_with(
        f"UInt{expected_datatype}", wait=True
    )
